# Copyright 2006-2009 Scott Horowitz <stonecrest@gmail.com>
# Copyright 2009-2014 Jonathan Ballet <jon@multani.info>
#
# This file is part of Sonata.
#
# Sonata is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Sonata is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Sonata.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os
import importlib.resources
import sys

from gi.repository import Gtk, Gdk


logger = logging.getLogger(__name__)


def builder(ui_file, relative_to='.'):
    builder = Gtk.Builder()
    builder.set_translation_domain('sonata')
    ui_path = str(importlib.resources.files('sonata') / os.path.join(relative_to, 'ui', ui_file + ".glade"))
    logger.debug('Loading %s', ui_path)
    builder.add_from_file(ui_path)

    return builder

def css_provider(css_file, relative_to='.'):
    provider = Gtk.CssProvider()
    css_path = str(importlib.resources.files('sonata') / os.path.join(relative_to, 'ui', css_file + ".css"))
    logger.debug('Loading %s', css_path)
    provider.load_from_path(css_path)
    screen = Gdk.Screen.get_default()
    context = Gtk.StyleContext()
    context.add_provider_for_screen(screen, provider,
                                    Gtk.STYLE_PROVIDER_PRIORITY_USER)

    return provider

def builder_string(ui_file, relative_to='.'):
    ui_path = str(importlib.resources.files('sonata') / os.path.join(relative_to, 'ui', ui_file + ".glade"))
    logger.debug('Loading %s', ui_path)

    with open(ui_path, 'r') as builder_file:
        string = builder_file.read()
    return string

def show_msg(owner, message, title, role, buttons, default=None, response_cb=None):
    is_button_list = hasattr(buttons, '__getitem__')
    if not is_button_list:
        messagedialog = Gtk.MessageDialog(owner, Gtk.DialogFlags.MODAL|Gtk.DialogFlags.DESTROY_WITH_PARENT, Gtk.MessageType.WARNING, buttons, message)
    else:
        messagedialog = Gtk.MessageDialog(owner, Gtk.DialogFlags.MODAL|Gtk.DialogFlags.DESTROY_WITH_PARENT, Gtk.MessageType.WARNING, message_format=message)
        i = 0
        while i < len(buttons):
            messagedialog.add_button(buttons[i], buttons[i+1])
            i += 2
    messagedialog.set_title(title)
    messagedialog.set_role(role)
    if default is not None:
        messagedialog.set_default_response(default)
    if response_cb:
        messagedialog.connect("response", response_cb)
    response = messagedialog.run()
    value = response
    messagedialog.destroy()
    return value

def dialog_destroy(dialog_widget, _response_id):
    dialog_widget.destroy()

def show(widget):
    widget.set_no_show_all(False)
    widget.show_all()

def hide(widget):
    #widget.hide_all()
    widget.hide()
    widget.set_no_show_all(True)

def quote_label(label_value):
    """Quote the content of a label so that it's safe to display."""

    # Don't inadvertently create accelerators if the value contains a "_"
    result = label_value.replace("_", "__")

    return result

def change_cursor(cursortype):
    for w in Gtk.Window.list_toplevels():
        gdk_window = w.get_window()
        # some toplevel windows have no drawing area
        if gdk_window != None:
            gdk_window.set_cursor(cursortype)

def set_entry_invalid(editable):
    color = Gdk.RGBA()
    color.parse("red")
    editable.override_color(Gtk.StateFlags.NORMAL, color)

def reset_entry_marking(editable):
    editable.override_color(Gtk.StateFlags.NORMAL, None)


def icon(name):
    return Gtk.IconFactory.lookup_default(name)



# Copyright 2006-2009 Scott Horowitz <stonecrest@gmail.com>
# Copyright 2009-2014 Jonathan Ballet <jon@multani.info>
#
# This file is part of Sonata.
#
# Sonata is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Sonata is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Sonata.  If not, see <http://www.gnu.org/licenses/>.

import os
from subprocess import Popen, PIPE

try:
    from sonata.genversion import VERSION
    build_ver = VERSION
except ImportError:
    build_ver = None

# Should be the most recent release
default_version = "v1.7.3"

def _version():
    '''Get the version number of the sources

    First check the build generated file, fallback to git describe if this is
    not a build, finally fallback to the default most recent release.
    '''
    if build_ver:
        version = build_ver
    else:
        try:
            dir = os.path.dirname(__file__)
            version = Popen(["git", "describe", "--abbrev=4", "HEAD"],
                             cwd=dir, stdout=PIPE,
                             stderr=PIPE).communicate()[0].decode('utf-8')
            if not version:
                raise OSError
        except OSError:
            version = default_version
    return version.strip()[1:]

version = _version()



# Copyright 2006-2009 Scott Horowitz <stonecrest@gmail.com>
# Copyright 2009-2014 Jonathan Ballet <jon@multani.info>
#
# This file is part of Sonata.
#
# Sonata is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Sonata is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Sonata.  If not, see <http://www.gnu.org/licenses/>.

import gettext
import locale
import logging
import os
from optparse import OptionParser
import sys

from sonata import config, misc, mpdhelper as mpdh
from sonata.version import version

# the mpd commands need a connection to server and exit without gui
mpd_cmds = ["play", "pause", "stop", "next", "prev", "pp", "info",
            "status", "repeat", "random"]


class Args:

    def __init__(self):
        self.logger = logging.getLogger(__name__)
        self.skip_gui = False
        self.start_visibility = None
        self.start_shell = False

    def parse(self, argv):
        """Parse the command line arguments.

        Separates options and arguments from the given argument list,
        checks their validity."""

        # toggle and popup and fullscreen need d-bus and don't always need gui
        # version and help don't need anything and exit without gui
        # hidden and visible are only applicable when gui is launched
        # profile and no-start don't need anything
        _usage = "\n".join((_("%prog [OPTION]... [COMMAND]...") + "\n",
         _("Commands:"),
        "  play            %s" % _("play song in playlist"),
        "  pause           %s" % _("pause currently playing song"),
        "  stop            %s" % _("stop currently playing song"),
        "  next            %s" % _("play next song in playlist"),
        "  prev            %s" % _("play previous song in playlist"),
        "  pp              %s" % _("toggle play/pause; plays if stopped"),
        "  repeat          %s" % _("toggle repeat mode"),
        "  random          %s" % _("toggle random mode"),
        "  info            %s" % _("display current song info"),
        "  status          %s" % _("display MPD status"),
        ))
        _version = "%prog " + version

        parser = OptionParser(usage=_usage, version=_version)
        parser.add_option("-p", "--popup", dest="popup",
                  action="store_true",
                  help=_("popup song notification (requires D-Bus)"))
        parser.add_option("-t", "--toggle", dest="toggle",
                  action="store_true",
                  help=_("toggles whether the app is minimized to the tray or visible (requires D-Bus)"))
        parser.add_option("-f", "--fullscreen", dest="fullscreen",
                  action="store_true",
                  help=_("go fullscreen (requires D-Bus)"))
        parser.add_option("-n", "--no-start", dest="start",
                  action="store_false",
                  help=_("don't start app if D-Bus commands fail"))
        parser.add_option("--hidden", dest="start_visibility",
                  action="store_false",
                  help=_("start app hidden (requires systray)"))
        parser.add_option("--visible", dest="start_visibility",
                  action="store_true",
                  help=_("start app visible (requires systray)"))
        parser.add_option("--profile", dest="profile", metavar="NUM",
                  help=_("start with profile NUM"), type=int)
        parser.add_option("--shell", dest="start_shell",
                  action="store_true",
                  help=_("start app with a interactive (ipython) shell"))
        parser.add_option("-v", "--verbose", dest="log_level",
                          action="append_const", const=-10,
                          help=_("Increase log verbosity"))
        parser.add_option("-q", "--quiet", dest="log_level",
                          action="append_const", const=10,
                          help=_("Decrease log verbosity"))
        parser.set_defaults(log_level=[logging.root.level])

        options, self.cmds = parser.parse_args(argv[1:])

        # Update default log level
        logging.root.setLevel(sum(options.log_level))

        if logging.root.level >= logging.DEBUG:
            # Only display MPD logs if we area really trying to debug hard
            # (-vvv)
            logging.getLogger('mpd').setLevel(logging.WARNING)

        if options.toggle:
            options.start_visibility = True
        if options.popup or options.fullscreen and options.start_visibility is None:
            options.start_visibility = False
        self.start_visibility = options.start_visibility
        self.arg_profile = options.profile
        self.start_shell = options.start_shell

        for cmd in self.cmds:
            if cmd in mpd_cmds:
                self.skip_gui = True
            else:
                parser.error(_("unknown command %s") % cmd)

        if options.toggle or options.popup or options.fullscreen:
            from sonata import dbus_plugin as dbus
            if not dbus.using_dbus():
                self.logger.critical(
                    _("toggle and popup options require D-Bus.  Aborting."))
                sys.exit(1)

            dbus.execute_remote_commands(options.toggle,
                             options.popup,
                             options.fullscreen,
                             options.start)

    def execute_cmds(self):
        """If arguments were passed, perform action on them."""
        if self.cmds:
            main = CliMain(self)
            main.mpd_connect()
            for cmd in self.cmds:
                main.execute_cmd(cmd)
            sys.exit()

    def apply_profile_arg(self, config):
        if self.arg_profile:
            a = self.arg_profile
            if a > 0 and a <= len(config.profile_names):
                config.profile_num = a-1
                self.logger.info(_("Starting Sonata with profile %s...") %
                                 config.profile_names[config.profile_num])
            else:
                self.logger.critical(
                    _("%d is not an available profile number.") % a)
                self.logger.critical(
                    _("Profile numbers must be between 1 and %d.") %
                    len(config.profile_names))
                sys.exit(1)


class CliMain:

    def __init__(self, args):
        global os, mpd, misc, config, mpdh
        import os
        import mpd
        from sonata import config, misc, mpdhelper as mpdh

        # This should disable processing for nearly all logging messages, as
        # long as no other handlers are configured for special loggers.
        # Note that using the "-v" command line argument is mostly useless in
        # this case...
        logging.root.setLevel(logging.CRITICAL)

        self.logger = logging.getLogger(__name__)
        self.config = config.Config(_('Default Profile'),
                                    _("by") + " %A " + _("from") + " %B")
        self.config.settings_load_real()
        args.apply_profile_arg(self.config)

        self.mpd = mpdh.MPDClient()

    def mpd_connect(self):
        host, port, password = misc.mpd_env_vars()
        if not host:
            host = self.config.host[self.config.profile_num]
        if not port:
            port = self.config.port[self.config.profile_num]
        if not password:
            password = self.config.password[self.config.profile_num]

        self.mpd.connect(host, port)
        if password:
            self.mpd.password(password)

    def execute_cmd(self, cmd):
        self.status = self.mpd.status()
        if not self.status:
            self.logger.critical(_(
                'Unable to connect to MPD.\nPlease check your Sonata '
                'preferences or MPD_HOST/MPD_PORT environment variables.'))
            sys.exit(1)

        self.songinfo = self.mpd.currentsong()
        getattr(self, "_execute_%s" % cmd)()

    def _execute_play(self):
        self.mpd.play()

    def _execute_pause(self):
        self.mpd.pause(1)

    def _execute_stop(self):
        self.mpd.stop()

    def _execute_next(self):
        self.mpd.next()

    def _execute_prev(self):
        self.mpd.previous()

    def _execute_random(self):
        opt = int(not int(self.status["random"]))
        self.mpd.random(opt)

    def _execute_repeat(self):
        opt = int(not int(self.status["repeat"]))
        self.mpd.repeat(opt)

    def _execute_pp(self):
        if self.status['state'] in ['play']:
            self.mpd.pause(1)
        elif self.status['state'] in ['pause', 'stop']:
            self.mpd.play()

    def _execute_info(self):
        if self.status['state'] in ['play', 'pause']:
            cmds = [
                (_("Title"), 'title'),
                (_("Artist"), 'artist'),
                (_("Album"), 'album'),
                (_("Date"), 'date'),
                (_("Track"), 'track'),
                (_("Genre"), 'genre'),
                (_("File"), 'file'),
            ]
            for pretty, key in cmds:
                value = str(self.songinfo.get(key))
                if key == 'track':
                    value = value.zfill(2)
                # XXX this could fail, if not all characters are supported by
                # os.stdout.encoding
                print("%s: %s" % (pretty, value))

            at, _length = [int(c) for c in self.status['time'].split(':')]
            at_time = misc.convert_time(at)
            try:
                time = misc.convert_time(self.songinfo.time)
                print("%s: %s/%s" % (_("Time"), at_time, time))
            except:
                print("%s: %s" % (_("Time"), at_time))
            print("%s: %s" % (_("Bitrate"),
                      self.status.get('bitrate', '')))
        else:
            print(_("MPD stopped"))

    def _execute_status(self):
        state_map = {
                'play': _("Playing"),
                'pause': _("Paused"),
                'stop': _("Stopped")}
        print("%s: %s" % (_("State"), state_map[self.status['state']]))

        print("%s %s" % (_("Repeat:"), _("On") \
                         if self.status['repeat'] == '1' else _("Off")))
        print("%s %s" % (_("Random:"), _("On") \
                         if self.status['random'] == '1' else _("Off")))
        print("%s: %s/100" % (_("Volume"), self.status['volume']))
        print("%s: %s %s" % (_('Crossfade'), self.status['xfade'],
                             ngettext('second', 'seconds',
                                      int(self.status['xfade']))))



# Copyright 2006-2009 Scott Horowitz <stonecrest@gmail.com>
# Copyright 2009-2014 Jonathan Ballet <jon@multani.info>
#
# This file is part of Sonata.
#
# Sonata is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Sonata is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Sonata.  If not, see <http://www.gnu.org/licenses/>.

import sys
import os
import locale
import logging
import threading

from gi.repository import Gtk, Pango, Gdk, GdkPixbuf, GLib

from sonata import ui, misc, consts, mpdhelper as mpdh, img
from sonata.pluginsystem import pluginsystem


def target_lyrics_filename(cfg, artist, title, song_dir, force_location=None):
    """Get the filename of the lyrics of a song"""

    lyrics_loc = force_location if force_location else cfg.lyrics_location

    if song_dir is not None:
        song_dir.replace('%', '%%')

    music_dir = cfg.musicdir[cfg.profile_num].replace('%', '%%')
    pattern1 = "%s-%s.txt"
    pattern2 = "%s - %s.txt"

    # Note: *_ALT searching is for compatibility with other mpd clients
    # (like ncmpcpp):
    file_map = {
        consts.LYRICS_LOCATION_HOME: ("~/.lyrics", pattern1),
        consts.LYRICS_LOCATION_PATH: (music_dir, song_dir, pattern1),
        consts.LYRICS_LOCATION_HOME_ALT: ("~/.lyrics", pattern2),
        consts.LYRICS_LOCATION_PATH_ALT: (music_dir, song_dir, pattern2),
    }

    file_path = os.path.join(*file_map[lyrics_loc])
    file_path = os.path.expanduser(file_path) % (artist, title)

    return misc.file_exists_insensitive(file_path)


class Info:

    def __init__(self, config, linkcolor, on_link_click_cb,
                 get_playing_song, TAB_INFO, on_image_activate,
                 on_image_motion_cb, on_image_drop_cb,
                 album_return_artist_and_tracks, add_tab):
        self.logger = logging.getLogger(__name__)
        self.config = config
        self.linkcolor = linkcolor
        self.on_link_click_cb = on_link_click_cb
        self.get_playing_song = get_playing_song
        self.album_return_artist_and_tracks = album_return_artist_and_tracks

        self.last_bitrate = None

        self.info_boxes_in_more = None
        self._editlabel = None
        self.info_left_label = None
        self.info_lyrics = None
        self._morelabel = None
        self._searchlabel = None

        self.lyrics_text = None
        self.album_text = None

        self.active = False

        # Info tab
        self.builder = ui.builder('info')
        self.css_provider = ui.css_provider('info')
        self.info_area = self.builder.get_object('info_page_scrolledwindow')
        self.tab_label_widget = self.builder.get_object('info_tab_eventbox')
        tab_label = self.builder.get_object('info_tab_label')
        tab_label.set_text(TAB_INFO)
        self.tab = add_tab(self.info_area, self.tab_label_widget,
                           TAB_INFO, self.info_area)

        self._imagebox = self.builder.get_object('info_page_song_eventbox')
        self.image = self.builder.get_object('info_page_song_image')
        self.image.set_from_icon_set(ui.icon('sonata-cd-large'), -1)

        self._imagebox.drag_dest_set(Gtk.DestDefaults.HIGHLIGHT |
                                     Gtk.DestDefaults.DROP,
                                     [Gtk.TargetEntry.new("text/uri-list", 0, 80),
                                      Gtk.TargetEntry.new("text/plain", 0, 80)],
                                      Gdk.DragAction.DEFAULT)
        self._imagebox.connect('button_press_event', on_image_activate)
        self._imagebox.connect('drag_motion', on_image_motion_cb)
        self._imagebox.connect('drag_data_received', on_image_drop_cb)

        self._widgets_initialize()

    def _widgets_initialize(self):
        self._widgets_song()
        self._widgets_lyrics()
        self._widgets_album()

        # Finish..
        if not self.config.show_lyrics:
            ui.hide(self.info_lyrics)
        if not self.config.show_covers:
            ui.hide(self._imagebox)

    def _widgets_song(self):
        self.info_song = self.builder.get_object('info_page_song_expander')
        self.info_song.set_expanded(self.config.info_song_expanded)
        self.info_song_grid = self.builder.get_object('info_page_song_grid')

        self.info_song.connect("activate", self._expanded, "song")

        self.info_labels = {}
        names = ('title', 'artist', 'album', 'date',
                 'track', 'genre', 'file', 'bitrate')
        for name in names:
            self.info_labels[name] = self.builder.get_object(
                'info_song_{}_label'.format(name))
        artist_eventbox = self.builder.get_object('info_song_artist_eventbox')
        album_eventbox = self.builder.get_object('info_song_album_eventbox')
        links = {
            'artist': artist_eventbox,
            'album': album_eventbox,}
        for name, widget in links.items():
            self._apply_link_signals(widget, name)

        file_label = self.builder.get_object('info_song_file_label_label')
        bitrate_label = self.builder.get_object('info_song_bitrate_label_label')
        self.info_boxes_in_more = {
            'values': (self.info_labels['file'], self.info_labels['bitrate'],),
            'labels': (file_label, bitrate_label,),}

        self._morelabel = self.builder.get_object('info_song_links_more_label')
        self.toggle_more()
        moreevbox = self.builder.get_object('info_song_links_more_eventbox')
        self._apply_link_signals(moreevbox, 'more')
        self._editlabel = self.builder.get_object('info_song_links_edit_label')
        editevbox = self.builder.get_object('info_song_links_edit_eventbox')
        self._apply_link_signals(editevbox, 'edit')

    def _widgets_lyrics(self):
        self.info_lyrics = self.builder.get_object('info_page_lyrics_expander')
        self.info_lyrics.set_expanded(self.config.info_lyrics_expanded)
        self.info_lyrics.connect("activate", self._expanded, "lyrics")
        self.lyrics_scrolledwindow = self.builder.get_object(
            'info_page_lyrics_scrolledwindow')
        self.lyrics_text = self.builder.get_object('info_page_lyrics_textview')
        self._populate_lyrics_tag_table()
        self._searchlabel = self.builder.get_object('info_page_lyrics_search')
        search_eventbox = self.builder.get_object(
            'info_page_lyrics_search_eventbox')
        self._apply_link_signals(search_eventbox, 'search')

    def _widgets_album(self):
        self.info_album = self.builder.get_object('info_page_album_expander')
        self.info_album.set_expanded(self.config.info_album_expanded)
        self.album_text = self.builder.get_object('info_page_album_textview')
        self.album_scrolledwindow = self.builder.get_object(
            'info_page_album_scrolledwindow')
        self.info_album.connect("activate", self._expanded, "album")

    def get_widgets(self):
        return self.info_area

    def get_info_imagebox(self):
        return self._imagebox

    def show_lyrics_updated(self):
        func = "show" if self.config.show_lyrics else "hide"
        getattr(ui, func)(self.info_lyrics)

    def _apply_link_signals(self, widget, linktype):
        widget.connect("enter-notify-event", self.on_link_enter)
        widget.connect("leave-notify-event", self.on_link_leave)
        widget.connect("button-press-event", self.on_link_click, linktype)

    def on_link_enter(self, widget, _event):
        ui.change_cursor(Gdk.Cursor.new(Gdk.CursorType.HAND2))

    def on_link_leave(self, _widget, _event):
        ui.change_cursor(None)

    def toggle_more(self):
        if self.config.info_song_more:
            text = _("hide")
            func = "show"
        else:
            text = _("more")
            func = "hide"
        text = "({})".format(text)
        self._morelabel.set_text(text)

        func = getattr(ui, func)
        for widget in self.info_boxes_in_more['labels']:
            func(widget)

        if not self.active and self.config.info_song_more:
            return

        for widget in self.info_boxes_in_more['values']:
            func(widget)

    def on_link_click(self, _widget, _event, linktype):
        if linktype == 'more':
            self.config.info_song_more = not self.config.info_song_more
            self.toggle_more()
        else:
            self.on_link_click_cb(linktype)

    def _expanded(self, expander, infotype):
        setattr(self.config, "info_%s_expanded" % infotype,
                not expander.get_expanded())

    def clear_info(self):
        """Clear the info widgets of any information"""
        for label in self.info_labels.values():
            label.hide()
        self._searchlabel.hide()
        self._show_lyrics(None, None)
        self.album_text.get_buffer().set_text("")
        self.last_bitrate = ""

    def update(self, playing_or_paused, newbitrate, songinfo, update_all):
        # update_all = True means that every tag should update. This is
        # only the case on song and status changes. Otherwise we only
        # want to update the minimum number of widgets so the user can
        # do things like select label text.
        if not playing_or_paused:
            self.clear_info()
            self.active = False
            return

        self.active = True

        for label in self.info_labels.values():
            if self.config.info_song_more or \
               not label in self.info_boxes_in_more['values']:
                label.show()
        bitratelabel = self.info_labels['bitrate']
        if self.last_bitrate != newbitrate:
            bitratelabel.set_text(newbitrate)
            self.last_bitrate = newbitrate

        if update_all:
            for func in ["song", "album", "lyrics"]:
                getattr(self, "_update_%s" % func)(songinfo)

    def _update_song(self, songinfo):
        artistlabel = self.info_labels['artist']
        tracklabel = self.info_labels['track']
        albumlabel = self.info_labels['album']
        filelabel = self.info_labels['file']

        for name in ['title', 'date', 'genre']:
            label = self.info_labels[name]
            label.set_text(songinfo.get(name, ''))

        tracklabel.set_text(str(songinfo.track))
        artistlabel.set_text(misc.escape_html(songinfo.artist))
        albumlabel.set_text(misc.escape_html(songinfo.album))

        path = os.path.join(self.config.current_musicdir, songinfo.file)
        if os.path.exists(path):
            filelabel.set_text(path)
            self._editlabel.show()
        else:
            filelabel.set_text(songinfo.file)
            self._editlabel.hide()

    def _update_album(self, songinfo):
        if 'album' not in songinfo:
            self.album_text.get_buffer().set_text(_("Album name not set."))
            return

        artist, tracks = self.album_return_artist_and_tracks()
        albuminfo = _("Album info not found.")

        if tracks:
            tracks.sort(key=lambda x: x.track)
            playtime = 0
            tracklist = []
            for t in tracks:
                playtime += t.time
                tracklist.append("%s. %s" % (
                    str(t.track).zfill(2),
                    t.get('title', os.path.basename(t.file))))

            album = songinfo.album
            year = songinfo.date
            playtime = misc.convert_time(playtime)
            albuminfo = "\n".join(i for i in (album, artist, year,
                              playtime) if i)
            albuminfo += "\n\n"
            albuminfo += "\n".join(t for t in tracklist)

        self.album_text.get_buffer().set_text(albuminfo)

    def _update_lyrics(self, songinfo):
        if self.config.show_lyrics:
            if 'artist' in songinfo and 'title' in songinfo:
                self.get_lyrics_start(songinfo.artist, songinfo.title,
                                      songinfo.artist, songinfo.title,
                                      os.path.dirname(songinfo.file))
            else:
                self._show_lyrics(None, None, error=_(('Artist or song title '
                                                       'not set.')))

    def _check_for_local_lyrics(self, artist, title, song_dir):
        locations = [
            consts.LYRICS_LOCATION_HOME,
            consts.LYRICS_LOCATION_PATH,
            consts.LYRICS_LOCATION_HOME_ALT,
            consts.LYRICS_LOCATION_PATH_ALT]
        for location in locations:
            filename = target_lyrics_filename(self.config, artist, title,
                                              song_dir, location)
            if os.path.exists(filename):
                return filename

    def get_lyrics_start(self, search_artist, search_title, filename_artist,
                         filename_title, song_dir, force_fetch=False):
        filename_artist = misc.strip_all_slashes(filename_artist)
        filename_title = misc.strip_all_slashes(filename_title)
        filename = self._check_for_local_lyrics(filename_artist,
                                                filename_title, song_dir)
        lyrics = ""
        if filename:
            # If the lyrics only contain "not found", delete the file and try
            # to fetch new lyrics. If there is a bug in Sonata/SZI/LyricWiki
            # that prevents lyrics from being found, storing the "not found"
            # will prevent a future release from correctly fetching the lyrics.
            try:
                with open(filename, 'r', encoding="utf-8") as f:
                    lyrics = f.read()
            except IOError:
                pass

            if lyrics == _("Lyrics not found"):
                force_fetch = True

        if force_fetch:
            # Remove all lyrics for this song
            while filename is not None:
                filename = self._check_for_local_lyrics(filename_artist,
                                                        filename_title,
                                                        song_dir)
                if filename is not None:
                    misc.remove_file(filename)

        if filename:
            # Re-use lyrics from file:
            try:
                with open(filename, 'r', encoding="utf-8") as f:
                    lyrics = f.read()
            except IOError:
                pass
            # Strip artist - title line from file if it exists, since we
            # now have that information visible elsewhere.
            header = "%s - %s\n\n" % (filename_artist, filename_title)
            if lyrics[:len(header)] == header:
                lyrics = lyrics[len(header):]
            self._show_lyrics(filename_artist, filename_title, lyrics=lyrics)
        else:
            def communicate(artist_then, title_then, lyrics=None, error=None):
                """Schedule actions from the plugin thread into the main
                thread"""
                GLib.idle_add(self._show_lyrics, artist_then, title_then,
                              lyrics, error)

            # Fetch lyrics from plugins.
            thread = threading.Thread(
                name="LyricsFetcher",
                target=FetchLyricsWorker,
                args=(self.config, communicate,
                      search_artist, search_title, song_dir))
            thread.start()

    def _show_lyrics(self, artist_then, title_then, lyrics=None, error=None):
        # For error messages where there is no appropriate info:
        if not artist_then or not title_then:
            self._searchlabel.hide()
            if error:
                self.lyrics_text.get_buffer().set_text(error)
            elif lyrics:
                self.lyrics_text.get_buffer().set_text(lyrics)
            else:
                self.lyrics_text.get_buffer().set_text("")
            return

        # Verify that we are displaying the correct lyrics:
        songinfo = self.get_playing_song()
        if not songinfo:
            return
        artist_now = misc.strip_all_slashes(songinfo.artist)
        title_now = misc.strip_all_slashes(songinfo.title)
        if artist_now == artist_then and title_now == title_then:
            self._searchlabel.show()
            if error:
                self.lyrics_text.get_buffer().set_text(error)
            elif lyrics:
                self._set_lyrics(lyrics)
            else:
                self.lyrics_text.get_buffer().set_text("")

    def _set_lyrics(self, lyrics):
        if lyrics is None:
            return

        lyrics_buf = self.lyrics_text.get_buffer()
        lyrics_buf.set_text('')

        # pango needs only ampersand to be escaped
        lyrics = misc.unescape_html(lyrics).replace('&', '&amp;')

        try:
            attr_list, plain_text, accel_marker = Pango.parse_markup(lyrics)
        except:
            # failed to parse, use lyrics as it is
            lyrics_buf.set_text(lyrics)
            return

        attr_iter = attr_list.get_iterator()

        while True:
            range = attr_iter.range()
            font = attr_iter.get_font()[0]
            text = plain_text[range[0]:range[1]]

            tags = []
            if font.get_weight() == Pango.Weight.BOLD:
                tags.append('bold')
            if font.get_style() == Pango.Style.ITALIC:
                tags.append('italic')

            if tags:
                lyrics_buf.insert_with_tags_by_name(lyrics_buf.get_end_iter(),
                                                    text, *tags)
            else:
                lyrics_buf.insert(lyrics_buf.get_end_iter(), text)

            if not attr_iter.next():
                break

    def _populate_lyrics_tag_table(self):
        tag_table = self.lyrics_text.get_buffer().get_tag_table()

        bold_tag = Gtk.TextTag.new('bold')
        bold_tag.set_property('weight', Pango.Weight.BOLD)
        tag_table.add(bold_tag)

        italic_tag = Gtk.TextTag.new('italic')
        italic_tag.set_property('style', Pango.Style.ITALIC)
        tag_table.add(italic_tag)

    def on_artwork_changed(self, artwork_obj, pixbuf):
        if self._imagebox.get_size_request()[0] == -1:
            notebook_width = self.info_song_grid.get_parent().get_allocation().width
            grid = self.info_song_grid
            grid_allocation = grid.get_allocation()
            grid_height = grid_allocation.height
            grid_width = grid.get_preferred_width_for_height(grid_height)[0]
            fullwidth = notebook_width - (grid_width + 120)
            new_width = max(fullwidth, 150)
        else:
            new_width = 150

        (pix2, w, h) = img.get_pixbuf_of_size(pixbuf, new_width)
        pix2 = img.do_style_cover(self.config, pix2, w, h)
        pix2 = img.pixbuf_add_border(pix2)

        self.image.set_from_pixbuf(pix2)
        del pix2
        del pixbuf

    def on_artwork_reset(self, artwork_obj):
        self.image.set_from_icon_set(ui.icon('sonata-cd-large'), -1)


class FetchLyricsWorker:
    """Thread worker to fetch lyrics for a song

    This must use the `communicate` method to commuicate with the main thread to
    update the GUI.
    """

    def __init__(self, config, communicate, search_artist, search_title, song_dir):
        self.config = config
        self.communicate = communicate
        self.logger = logging.getLogger(__name__)

        self.fetch_lyrics(search_artist, search_title, song_dir)

    def fetch_lyrics(self, search_artist, search_title, song_dir):
        # Homogenize search patterns, so plugins don't have to do it.
        search_artist = str(search_artist).title()
        search_title = str(search_title).title()

        lyrics_fetchers = pluginsystem.get('lyrics_fetching')
        if lyrics_fetchers:
            self.logger.info("Looking for lyrics for %r - %r...",
                             search_artist, search_title)
            self.communicate(search_artist, search_title,
                             lyrics=_("Fetching lyrics..."))
            for plugin, get_lyrics in lyrics_fetchers:
                try:
                    lyrics = get_lyrics(search_artist, search_title)
                except Exception as e:
                    if self.logger.isEnabledFor(logging.DEBUG):
                        # Be more verbose if we want something verbose
                        log = self.logger.exception
                    else:
                        log = self.logger.warning

                    log("Plugin %s: unable to fetch lyrics (%s)",
                        plugin.name, e)
                    continue

                if lyrics:
                    self.logger.info("Lyrics for %r - %r fetched by %r plugin.",
                                     search_artist, search_title, plugin.name)
                    self.get_lyrics_response(search_artist, search_title,
                                             song_dir, lyrics=lyrics)
                    return
            msg = _("Lyrics not found.")
        else:
            self.logger.info("Can't look for lyrics, no plugin enabled.")
            msg = _("No lyrics plug-in enabled.")

        self.communicate(search_artist, search_title, lyrics=msg)

    def get_lyrics_response(self, artist_then, title_then, song_dir,
                lyrics=None, error=None):
        if lyrics and not error:
            filename = target_lyrics_filename(self.config, artist_then,
                                              title_then, song_dir)
            # Save lyrics to file:
            misc.create_dir('~/.lyrics/')
            self.logger.info("Saving lyrics to: %s", filename)
            try:
                with open(filename, 'w', encoding="utf-8") as f:
                    lyrics = misc.unescape_html(lyrics)
                    f.write(lyrics)
            except IOError as e:
                self.logger.warning("Can't save lyrics to %s: %s", filename, e)

        self.communicate(artist_then, title_then, lyrics, error)



# $Id$

"""
~~~~~~~~~~~
PyScrobbler
~~~~~~~~~~~

Python Classes for AudioScrobbler
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
A set of python classes for interacting with the Last.fm AudioScrobbler
services.

AudioScrobblerQuery
    AudioScrobblerQuery is used to consume the web services that provide
    similar artists, charts, tags and so on

AudioScrobblerPost
    AudioScrobblerPost is used to update your personal play information for
    those of you (like me) who've been stoopid enough to write your own player.

AudioScrobblerQuery: AudioScrobbler Web Service Consumption
=============================================================

A set of classes for consuming the AudioScrobbler web services.
Currently we consume protocol 1.0, as defined at:
http://www.audioscrobbler.net/data/webservices

Basic usage
-----------
>>> lips = AudioScrobblerQuery(artist='The Flaming Lips')
>>> for artist in lips.similar():
...     print artist.name, artist.mbid
...

or

>>> for i in AudioScrobblerQuery(user='offmessage').topalbums()
...     print i.name, i.image.small
...

A certain number of the services take a querystring (for example a user's
weekly artist chart takes a start and end date).  These can be passed as
follows:

>>> offmessage = AudioScrobblerQuery(user='offmessage')
>>> for i in offmessage.weeklyartistchart(_from=1114965332, to=1115570132):
...     print i.name, i.chartposition
...

Note that we have had to prefix the ``from`` keyword with an underscore
(``_``). This is because ``from`` is a reserved word in Python and can't be
used as a parameter name.  Any parameter passed with a leading underscore will
have that underscore stripped (so ``_from`` is passed as ``from``, as ``_foo``
would be passed as ``foo``).

If an element has an attribute (for example the artist has an ``mbid`` in
/user/.../topalbums.xml) these can be accessed as a dictionary key of that
element:

>>> offmessage = AudioScrobblerQuery(user='offmessage')
>>> for i in offmessage.topalbums():
...     print i.name, i.artist, i.artist['mbid']
...

if you're not sure if an element has an attribute they can be accessed instead
using the ``get()`` method:

>>> offmessage = AudioScrobblerQuery(user='offmessage')
>>> for i in offmessage.topalbums():
...     print i.name, i.artist, i.artist.get('mbid', '')
...

More control
------------
There are 2 additional keyword arguments that you can offer at the point of
instantiation of the AudioScrobblerQuery object:

host
    If the lovely people at Last.fm change their server name, or if you set up
    some kind of local proxy of their services you can set the host from which
    you get the XML.

    Usage:
    >>> lips = AudioScrobblerQuery(artist='The Flaming Lips',
    ...                              host='audioscrobbler.isotoma.com')

    Default is ``'ws.audioscrobbler.com'``

version
    If another version of the web services is released, and this code still
    holds up you can set the protocol version at instantiation.  This is
    inserted into the request URL at the correct point (i.e.
    ws.audioscrobbler.com/**1.0**/artist/The+Flaming+Lips).

    Usage:
    >>> lips = AudioScrobberRequest(artist='The Flaming Lips', version='1.1')

    Default is ``'1.0'``

Caching
-------
Each instance of the ``AudioScrobblerQuery`` maintains its own internal cache
of requests, which are checked against the server's response.  If the server
returns a ``304`` (content not modified) we return the cached value, otherwise
we return the result that the server gave us.

This is all well and good, however at the time of writing the
``ws.audioscrobbler.com`` server does not support the ``304`` response for all
pages, so you may find yourself making requests that you are sure should be
cached, but in fact aren't.

For example:
http://ws.audioscrobbler.com/1.0/user/offmessage/topalbums.xml returns a
``304`` http://ws.audioscrobbler.com/1.0/artist/The+Flaming+Lips/similar.xml
does not

I have assumed that this is for a reason, and as such honour this behaviour.
If a page does not return a ``Last-Modified`` header we do not cache it.  Any
page that does return such a header (but not the ``Pragma: no-cache`` header)
gets cached.  Any page that does return a ``Last-Modified`` header gets exactly
what it sent to us sent back to it as the ``If-Modified-Since``.  This is a fit
of pragmatism based on the fact that calculating local dates and avoiding
clock disparaties is not worth the effort.

Also, note that the AudioScrobbler servers do not return the ``ETag`` header
and so we do not look for it or refer to it.

A last note
-----------
These classes are for consuming the AudioScrobbler web services.  As such each
of the classes here have no ``set``, ``__setitem__`` or ``__setattr__``
methods.  They are read only.  If you want to actually manipulate the
resulting XML you should access the underlying ElementTree_ instances directly
using either the ``raw()`` or ``element()`` methods provided.  You can then
make use of the innate power of those classes directly:

``raw()`` returns an ElementTree:

>>> tree = AudioScrobblerQuery(user='offmessage').topalbums().raw()

while ``element()`` returns an Element at the point within the tree that you
currently are:

>>> lips = AudioScrobblerQuery(artist='The Flaming Lips')
>>> elements = [ x.name.element() for x in lips.similar() ]

which is equivalent to:

>>> lips = AudioScrobberRequest(artist='The Flaming Lips')
>>> elements = lips.similar().raw().findall('*/name')

As an alternative you could also look at XMLTramp_, which is another XML
manipulator designed primarily for consuming web services (and used in the
FlickrClient_ script that gave me much inspiration for these classes).

Last but not least a mention should go to BeautifulSoup_, just because it is
the most powerful and flexible of this type of application that I've used.  It
specialises in badly formed XML/HTML though, and luckily is not required in
this instance.  That said, it's a powerful beast and you should have a look at
it the minute your requirements become more complex than this code supports.

.. _ElementTree: http://effbot.org/zone/element.htm
.. _XMLTramp: http://www.aaronsw.com/2002/xmltramp/
.. _BeautifulSoup: http://www.crummy.com/software/BeautifulSoup/


AudioScrobblerPost: Update Your Personal Play History
=====================================================

A set of classes for posting playlist information to your Last.fm account.
Please note that you will need your own account to use these classes.

Updated to work with version 1.2 of the Audioscrobbler protocol.

Some stuff about the time management and internal cache should go here.

Basic usage (outdated)
-----------
>>> track = dict(artist_name="The Flaming Lips",
                 song_title="The Yeah Yeah Yeah Song",
                 length=291,
                 date_played="2006-04-29 04:20:00",
                 album="At War With The Mystics",
                 mbid=""
                )
>>> post = AudioScrobblerPost(user='offmessage', password='mypasswd')
>>> post(**track)
>>>

More control
------------
There are a number of keywords that can be passed at the time of instantiation
of the ``AudioScrobblerPost`` class:

verbose
    Boolean.  If you pass ``verbose=True`` every log line is also printed.
    Useful for debugging and in interactive scripts.

client_name
    By default pyscrobbler uses the ``tst`` client name.  If you deploy these
    classes in an application with a wide user base it seems to be advisable to
    apply for your own client name.  If lots of people start to use these
    classes as is I will apply for a unique name for them.  If you have your
    own, pass it in here (as a string)

client_version
    Currently we use ``1.0``.  This is because we are using the ``tst`` client
    name (see above) and anything lower seems to raise an ``UPDATE`` message
    from the post server at handshake.

protocol_version
    By default we use ``1.1``.  These classes are not compatible with Protocol
    1.0.  It's unlikely they will be compatible with any later versions.  The
    changes between 1.0 and 1.1 were pretty major; I can't see they'd be any
    less major for the next version.

host
    By default we use ``post.audioscrobbler.com``.  If the server name changes
    (or you want to run tests) you can pass a different host name here.  In
    fact, the url that we authenticate against is built as
    ``"http://%s/?%s" % (host, params)`` so as long as your host string makes
    a valid url in that format, go for it.  Note the trailing slash between
    host and the query string.  This seems to be a requirement of urllib2. This
    means that your test auth script should be able to be called that way.

Putting them to use in real life
================================
Some advice on using these classes.

``timeout``
-----------
I've not done anything about the socket timeout on the running system.  It's
not really this code's place to do anything about it, but it is something that
probably needs sorting out, otherwise your scripts will hang for as long as
your timeout setting if there is a problem with the Last.fm servers (which
there appears to be fairly frequently with the submissions servers).

Generally, if you are on a slow link or one with poor latency you can up the
timeout so that your requests don't keep failing unnecessarily.  However, if
you're on a quick link you may want to lower it as low as 10 seconds, as the
wait can be agonising if you're making a few requests and the server is not
responding.

Thanks
======
I took a considerable amount of inspiration from Michael Campeotto's
FlickrClient_ script for AudioScrobblerQuery

I've taken even greater amounts of inspiration from Mike Bleyer's
iPodScrobbler_ script for AudioScrobblerPost.

.. _FlickrClient: http://micampe.it/things/flickrclient
.. _iPodScrobbler: http://www.hoc.net/mike/source/iPodScrobbler/


TODO
====
Mad leet reStructuredText epydoc mashup
    Get that sorted.

Example query script
    Actually write some example query scripts :)

Document use of the factory class
    AudioScrobbler


License
=======
Copyright 2011 Andy Theyers

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

__author__ = "Andy Theyers <andy@isotoma.com>"
__version__ = "$Revision$"[11:-2]
__docformat__ = "restructuredtext"


import datetime
import locale
import hashlib
import site
import logging
import sys
import time
import urllib.error, urllib.parse, urllib.request
import configparser
import os

from xml.etree import ElementTree

# This is lifted in the most part from iPodScrobbler (see docs above)
# Get the base local encoding
enc = 'ascii'
try:
    enc = site.encoding
except:
    pass
if enc == 'ascii' and locale.getpreferredencoding():
    enc = locale.getpreferredencoding()
# if we are on MacOSX, we default to UTF8
# because apples python reports 'ISO8859-1' as locale, but MacOSX uses utf8
if sys.platform == 'darwin':
    enc = 'utf8'

# AudioScrobblerQuery configuration settings
audioscrobbler_request_version = '1.0'
audioscrobbler_request_host = 'ws.audioscrobbler.com'

# AudioScrobblerPost configuration settings
audioscrobbler_post_version = '1.2' #updated
audioscrobbler_post_host = 'post.audioscrobbler.com'
client_name = 'sna'
pyscrobbler_version = '1.0.0.0' # This is set to 1.0 while we use
                             # client_name = u'tst' as we keep getting
                             # UPDATE responses with anything less.


def lines_from_utf8(lines):
    return [line.decode('utf-8') for line in lines]


class AudioScrobblerError(Exception):

    """
    The base AudioScrobbler error from which all our other exceptions derive
    """

    def __init__(self, message):
        """ Create a new AudioScrobblerError.

            :Parameters:
                - `message`: The message that will be displayed to the user
        """
        self.message = message

    def __repr__(self):
        msg = "%s: %s"
        return msg % (self.__class__.__name__, self.message,)

    def __str__(self):
        return self.__repr__()


class AudioScrobblerConnectionError(AudioScrobblerError):

    """
    The base network error, raise by invalid HTTP responses or failures to
    connect to the server specified (DNS, timeouts, etc.)
    """

    def __init__(self, type, code, message):
        self.type = type
        self.code = code
        self.message = message

    def __repr__(self):
        msg = "AudioScrobblerConnectionError: %s: %s %s"
        return msg % (self.type.upper(), self.code, self.message,)


class AudioScrobblerTypeError(AudioScrobblerError):

    """
    If we would normally raise a TypeError we raise this instead
    """
    pass


class AudioScrobblerPostUpdate(AudioScrobblerError):

    """
    If the POST server returns an ``UPDATE`` message we raise this exception.
    This is the only case when this exception is raised, allowing you to
    happily ignore it.
    """

    def __repr__(self):
        msg = "An update to your AudioScrobbler client is available at %s"
        return msg % (self.message,)


class AudioScrobblerPostFailed(AudioScrobblerError):

    """
    If the POST server returns an ``FAILED`` message we raise this exception.
    """

    def __repr__(self):
        msg = "Posting track to AudioScrobbler failed.  Reason given: %s"
        return msg % (self.message,)


class AudioScrobblerHandshakeError(AudioScrobblerError):

    """
    If we fail the handshake this is raised.  If you're running in a long
    running process you should pass on this error, as the system keeps a
    cache which ``flushcache()`` will clear for you when the server is back.
    """
    pass


class AudioScrobbler:

    """ Factory for Queries and Posts.  Holds configuration for the session """

    def __init__(self,
                 audioscrobbler_request_version=audioscrobbler_request_version,
                 audioscrobbler_post_version=audioscrobbler_post_version,
                 audioscrobbler_request_host=audioscrobbler_request_host,
                 audioscrobbler_post_host=audioscrobbler_post_host,
                 client_name=client_name,
                 client_version=pyscrobbler_version):

        self.logger = logging.getLogger(__name__)

        self.audioscrobbler_request_version = audioscrobbler_request_version
        self.audioscrobbler_post_version = audioscrobbler_post_version
        self.audioscrobbler_request_host = audioscrobbler_request_host
        self.audioscrobbler_post_host = audioscrobbler_post_host
        self.client_name = client_name
        self.client_version = pyscrobbler_version

    def query(self, **kwargs):

        """ Create a new AudioScrobblerQuery """

        if len(kwargs) != 1:
            raise TypeError("__init__() takes exactly 1 argument, %s "
                            "given" % (len(kwargs),))
        ret = AudioScrobblerQuery(version=self.audioscrobbler_request_version,
                                  host=self.audioscrobbler_request_host,
                                  **kwargs)
        return ret

    def post(self, username, md5_password, verbose=False):

        """ Create a new AudioScrobblerPost """

        ret = AudioScrobblerPost(username=username.encode('utf8'),
                         md5_password=md5_password.encode('utf8'),
                         host=self.audioscrobbler_post_host,
                         protocol_version=self.audioscrobbler_post_version,
                         client_name=self.client_name,
                         client_version=self.client_version,
                         verbose=verbose)
        return ret


class AudioScrobblerCache:

    def __init__(self, elemtree, last):
        self.elemtree = elemtree
        self.requestdate = last

    def created(self):
        return self.requestdate

    def gettree(self):
        return self.elemtree


class AudioScrobblerQuery:

    def __init__(self,
                 version=audioscrobbler_request_version,
                 host=audioscrobbler_request_host,
                 **kwargs):

        if len(kwargs) != 1:
            raise TypeError("__init__() takes exactly 1 audioscrobbler "
                            "request argument, %s given" % str(len(kwargs)))
        self.type = kwargs.keys()[0]
        self.param = str(kwargs[self.type])
        self.baseurl = 'http://%s/%s/%s/%s' % (host,
                                               version,
                                               urllib.parse.quote(self.type),
                                               urllib.parse.quote(self.param),
                                              )
        self._cache = {}

    def __getattr__(self, name):

        def method(_self=self, name=name, **params):
            # Build the URL
            url = '%s/%s.xml' % (_self.baseurl, urllib.parse.quote(name))
            if len(params) != 0:
                for key in params.keys():
                    # This little mess is required to get round the fact that
                    # 'from' is a reserved word and can't be passed as a
                    # parameter name.
                    if key.startswith('_'):
                        params[key[1:]] = params[key]
                        del params[key]
                url = '%s?%s' % (url, urllib.parse.urlencode(params))
            req = urllib.request.Request(url)

            # Check the cache
            cache = _self._cache
            if url in cache and cache[url].created() is not None:
                req.add_header('If-Modified-Since', cache[url].created())

            # Open the URL and test the response
            try:
                response = urllib.request.urlopen(url)
            except urllib.error.HTTPError as error:
                if error.code == 304:
                    return AudioScrobblerItem(cache[url].getroot(), _self, url)
                if error.code == 400:
                    raise AudioScrobblerConnectionError('ws', 400,
                                                        error.fp.read())
                raise AudioScrobblerConnectionError('http', error.code,
                                                    error.msg)
            except urllib.error.URLError as error:
                code = error.reason.args[0]
                message = error.reason.args[1]
                raise AudioScrobblerConnectionError('network', code, message)
            elemtree = ElementTree.ElementTree(file=response)
            if response.headers.get('pragma', None) != 'no-cache':
                last_modified = response.headers.get('last-modified', None)
            else:
                last_modified = None
            _self._cache[url] = AudioScrobblerCache(elemtree, last_modified)
            return AudioScrobblerItem(elemtree.getroot(), _self, url)
        return method

    def __repr__(self):
        return "AudioScrobblerQuery(%s='%s')" % (self.type, self.param)

    def __str__(self):
        return self.__repr__()


class AudioScrobblerItem:

    def __init__(self, element, parent, url=None):
        self._element = element
        self._parent = parent
        self.tag = element.tag
        self.text = element.text
        self._url = url
        if self._url is None:
            self._url = self._parent._url

    def __repr__(self):
        if isinstance(self._parent, AudioScrobblerQuery):
            return "<AudioScrobbler response from %s>" % (self._url,)
        return "<AudioScrobbler %s element at %s>" % (self.tag, id(self))

    def __str__(self):
        if self.text is None:
            return ''
        text = self.text
        try:
            retval = text.encode(enc)
        except AttributeError:
            retval = text
        return retval

    def __getattr__(self, name):
        result = self._element.findall(name)
        if len(result) == 0:
            raise AttributeError("AudioScrobbler %s element has no "
                                 "subelement '%s'" % (self.tag, name))
        ret = [AudioScrobblerItem(i, self) for i in result]
        if len(ret) == 1:
            return ret[0]
        return ret

    def __iter__(self):
        for i in self._element:
            yield AudioScrobblerItem(i, self)

    def __getitem__(self, i):
        return self._element.attrib[i]

    def get(self, i, default):
        if i in self._element.attrib:
            return self._element.attrib[i]
        else:
            return default

    def __getslice__(self, i, j):
        return [AudioScrobblerItem(x, self) for x in self._element[i:j]]

    def raw(self):

        def getparent(obj):
            if isinstance(obj._parent, AudioScrobblerQuery):
                return obj._parent
            return getparent(obj._parent)
        return getparent(self)._cache[self._url].gettree()

    def element(self):
        return self._element


class AudioScrobblerPost:

    """
    Provide the ability to post tracks played to a user's Last.fm
    account
    """
    #Updated to protocol version 1.2

    def __init__(self,
                 username='',
                 md5_password='',
                 client_name=client_name,
                 client_version=pyscrobbler_version,
                 protocol_version=audioscrobbler_post_version,
                 host=audioscrobbler_post_host,
                 verbose=False):

        # Capture the information passed for future use
        self.params = dict(username=username,
                           md5_password=md5_password,
                           client_name=client_name,
                           client_version=client_version,
                           protocol_version=protocol_version,
                           host=host)

        self.logger = logging.getLogger(__name__)
        if not verbose:
            self.logger.setLevel(logging.WARNING)

        self.auth_details = {}
        self.last_post = None
        self.interval = 0
        self.cache = []
        self.loglines = [] # TODO: it somehow duplicates with the logging module
        self.last_shake = None
        self.authenticated = False
        self.updateurl = None
        self.posturl = None
        self.npurl = None

    def auth(self):

        """ Authenticate against the server """

        if self.authenticated:
            return True

        timestamp = str(int(time.time())).encode('ascii')
        md5_password = self.params['md5_password']
        auth_token = hashlib.md5(md5_password + timestamp).hexdigest()

        p = {}
        p['hs'] = 'true'
        p['p'] = self.params['protocol_version']
        p['c'] = self.params['client_name']
        p['v'] = self.params['client_version']
        p['u'] = self.params['username']
        p['t'] = timestamp
        p['a'] = auth_token

        plist = [(k, urllib.parse.quote_plus(v).encode('utf-8'))
                 for k, v in p.items()]

        authparams = urllib.parse.urlencode(plist)
        url = 'http://%s/?%s' % (self.params['host'], authparams)
        req = urllib.request.Request(url)
        try:
            url_handle = urllib.request.urlopen(req)
        except urllib.error.HTTPError as error:
            self.authenticated = False
            raise AudioScrobblerConnectionError('http', error.code, error.msg)
        except urllib.error.URLError as error:
            self.authenticated = False
            code = error.reason#.args[0]
            message = error.reason#.args[1]
            raise AudioScrobblerConnectionError('network', code, message)

        response = lines_from_utf8(url_handle.readlines())
        if len(response) == 0:
            raise AudioScrobblerHandshakeError(('Got nothing back from '
                                                'the server'))

        username = self.params['username']
        md5_password = self.params['md5_password']

        # First we test the best and most likely case
        if response[0].startswith('OK'):
            answer = response[1].strip()
            self.auth_details['s'] = answer
            self.npurl = response[2].strip()
            self.posturl = response[3].strip()
            self.authenticated = True

        # Then the various failure states....
        elif response[0].startswith('BANNED'):
            self.authenticated = False
            msg = "this client version has been banned from the server."
            raise AudioScrobblerHandshakeError(msg)

        elif response[0].startswith('BADAUTH'):
            self.authenticated = False
            msg = "The authentication details provided were incorrect."
            raise AudioScrobblerHandshakeError(msg)

        elif response[0].startswith('BADTIME'):
            self.authenticated = False
            msg = ('The timestamp provided was not close enough to '
                   'the current time.')
            raise AudioScrobblerHandshakeError(msg)

        elif response[0].startswith('FAILED'):
            self.authenticated = False
            msg = response[0][6:]
            raise AudioScrobblerHandshakeError(msg)

        else:
            self.authenticated = False
            msg = "Unknown response from server: %r" % (response,)
            raise AudioScrobblerHandshakeError(msg)

    def posttrack(self,
                  artist_name,
                  song_title,
                  length,
                  date_played,
                  tracknumber='',
                  album='',
                  mbid='',
                  source='P'):

        """
        Add the track to the local cache, and try to post it to the server
        """

        self.addtrack(artist_name=artist_name,
                      song_title=song_title,
                      length=length,
                      date_played=date_played,
                      tracknumber=tracknumber,
                      album=album,
                      mbid=mbid,
                      source=source)
        self.post()

    __call__ = posttrack

    def addtrack(self,
                 artist_name,
                 song_title,
                 length,
                 date_played,
                 tracknumber='',
                 album='',
                 mbid='',
                 source='P'):

        """ Add a track to the local cache """

        # Quick sanity check on track length
        if type(length) == type(0):
            sane_length = length
        elif type(length) == type('') and length.isdigit():
            sane_length = int(length)
        else:
            sane_length = -1

        # If the track length is less than 30 move on
        if sane_length < 30:
            msg = ("Track '%s' by '%s' only %s seconds in length, so "
                   "not added" % (song_title, artist_name, sane_length))
            self.log(msg)
        # Otherwise build the track dictionary and add it to the local cache
        else:
            try:
                track = {'a[%s]': artist_name.encode('utf8'),
                        't[%s]': song_title.encode('utf8'),
                        'l[%s]': str(sane_length),
                        'i[%s]': date_played,
                        'b[%s]': album.encode('utf8'),
                        'm[%s]': mbid.encode('utf8'),
                        'r[%s]': '',
                        'n[%s]': str(tracknumber),
                        'o[%s]': source.encode('utf8'),
                        }
            except:
                track = {'a[%s]': artist_name,
                        't[%s]': song_title,
                        'l[%s]': str(sane_length),
                        'i[%s]': date_played,
                        'b[%s]': album,
                        'm[%s]': mbid.encode('utf8'),
                        'r[%s]': '',
                        'n[%s]': str(tracknumber),
                        'o[%s]': source.encode('utf8'),
                        }
            self.cache.append(track)

    def nowplaying(self,
                  artist_name,
                  song_title,
                  length='',
                  tracknumber='',
                  album='',
                  mbid=''):

        self.auth()

        p = {}
        p['s'] = self.auth_details['s']
        p['a'] = artist_name
        p['t'] = song_title
        p['b'] = album
        p['l'] = str(length)
        p['n'] = str(tracknumber)
        p['m'] = str(mbid)

        npdata = urllib.parse.urlencode(p).encode('utf-8')

        req = urllib.request.Request(url=self.npurl, data=npdata)

        try:
            url_handle = urllib.request.urlopen(req)
        except urllib.error.HTTPError as error:
            self.authenticated = False
            self.logger.error(
                "%s", AudioScrobblerConnectionError('http', error.code, error.msg))
            return
        except:
            code = '000'
            message = sys.exc_info()[1]
            self.logger.error(
                "%s", AudioScrobblerConnectionError('network', code, message))
            return

        response = lines_from_utf8(url_handle.readlines())
        if response[0].startswith('OK'):
            self.log("Now playing track updated ('%s' by '%s')." % (p['t'],
                                                                    p['a']))
        elif response[0].startswith('BADSESSION'):
            self.authenticated = False

    def post(self):

        """
        Post the tracks by popping the first ten off the cache and attempting
        to post them.
        """

        if len(self.cache) == 0:
            return
        if len(self.cache) > 10:
            number = 10
        else:
            number = len(self.cache)

        params = {}
        count = 0
        for track in self.cache[:number]:
            for k in track.keys():
                params[k % (count,)] = track[k]
            count += 1

        self.auth()
        params.update(self.auth_details)
        postdata = urllib.parse.urlencode(params).encode('utf-8')
        req = urllib.request.Request(url=self.posturl, data=postdata)

        now = datetime.datetime.utcnow()

        try:
            url_handle = urllib.request.urlopen(req)
        except urllib.error.HTTPError as error:
            raise AudioScrobblerConnectionError('http', error.code, error.msg)
        except urllib.error.URLError as error:
            args = getattr(error.reason, 'args', None)
            code = '000'
            message = str(error)
            if args is not None:
                if len(args) == 1:
                    message = error.reason.args[0]
                elif len(args) == 2:
                    code = error.reason.args[0]
                    message = error.reason.args[1]
            raise AudioScrobblerConnectionError('network', code, message)
        except:
            code = '000'
            message = sys.exc_info()[1]
            raise AudioScrobblerConnectionError('network', code, message)

        self.last_post = now
        response = lines_from_utf8(url_handle.readlines())

        # Test the various responses possibilities:
        if response[0].startswith('OK'):
            for song in self.cache[:number]:
                self.log("Uploaded track successfully ('%s' by '%s')." \
                         % (song['t[%s]'], song['a[%s]']))
            del self.cache[:number]
        elif response[0].startswith('BADSESSION'):
            self.log("Got BADSESSION.")
            self.authenticated = False
        elif response[0].startswith('FAILED'):
            reason = response[0][6:].strip()
            raise AudioScrobblerPostFailed(reason)
        else:
            msg = "Server returned something unexpected."
            raise AudioScrobblerPostFailed(msg)

    def flushcache(self):

        """ Post all the tracks in the cache """

        while len(self.cache) > 0:
            self.post()

    def savecache(self, filename):

        """ Save all tracks in the cache to a file """

        conf = configparser.ConfigParser()

        # Save each track in cache:
        count = 0
        for track in self.cache:
            conf.add_section('Track ' + str(count))
            for k in track.keys():
                conf.set('Track ' + str(count), k, track[k])
            count += 1

        try:
            conf.write(open(filename, 'w'))
        except IOError:
            pass

    def retrievecache(self, filename):

        """
        Retrieve all cached tracks from a file so that cached tracks
        are preserved across client restarts
        """

        if not os.path.isfile(filename):
            return

        conf = configparser.ConfigParser()
        try:
            conf.read(filename)
        except IOError:
            return

        # Retrieve each cached track from file:
        count = 0
        while conf.has_section('Track ' + str(count)):
            track = {}
            for key in ['a', 't', 'l', 'i', 'b', 'm', 'r', 'n', 'o']:
                if conf.has_option('Track ' + str(count), key + '[%s]'):
                    track[key + '[%s]'] = conf.get('Track %s' % (str(count),),
                                                   key + '[%s]')
            self.cache.append(track)
            count += 1

        # Cached tracks retrieved, delete file:
        os.remove(filename)

    def log(self, msg):

        """ Add a line to the log, print it if verbose """

        time = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        self.loglines.append("%s: %s" % (time, msg))
        self.logger.debug(msg)

    def getlog(self, clear=False):

        """ Return the entire log, clear it if requested """

        if clear:
            retval = self.loglines
            self.loglines = []
            return retval
        return self.loglines



# coding=utf-8
# Copyright 2006-2009 Scott Horowitz <stonecrest@gmail.com>
# Copyright 2009-2014 Jonathan Ballet <jon@multani.info>
#
# This file is part of Sonata.
#
# Sonata is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Sonata is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Sonata.  If not, see <http://www.gnu.org/licenses/>.

import gettext
import os

from gi.repository import Gtk, GdkPixbuf

from sonata import misc, ui

class About:

    def __init__(self, parent_window, config, version):
        self.parent_window = parent_window
        self.config = config
        self.version = version

        self.about_dialog = None
        self.shortcuts_dialog = None

    def about_close(self, _event, _data=None):
        if _data == Gtk.ResponseType.DELETE_EVENT or \
           _data == Gtk.ResponseType.CANCEL:
            self.about_dialog.hide()
        return True

    def shortcuts_close(self, _event, _data=None):
        if _data == Gtk.ResponseType.DELETE_EVENT or \
           _data == Gtk.ResponseType.CANCEL:
            self.shortcuts_dialog.hide()
        return True

    def shortcuts_close_click_cb(self, _button):
        self.shortcuts_dialog.hide()

    def _init_shortcuts_dialog(self):
        # define the shortcuts and their descriptions
        # these are all gettextable
        # Keep them here (not as XML) as they're more convenient this way
        mainshortcuts = \
                [["F1", _("About Sonata")],
                 ["F5", _("Preferences")],
                 ["F11", _("Fullscreen Artwork Mode")],
                 ["Alt-[1-5]", _("Switch to [1st-5th] tab")],
                 ["Alt-C", _("Connect to MPD")],
                 ["Alt-D", _("Disconnect from MPD")],
                 ["Alt-R", _("Randomize current playlist")],
                 ["Alt-Down", _("Expand player")],
                 ["Alt-Left", _("Switch to previous tab")],
                 ["Alt-Right", _("Switch to next tab")],
                 ["Alt-Up", _("Collapse player")],
                 ["Ctrl-H", _("Search library")],
                 ["Ctrl-Q", _("Quit")],
                 ["Ctrl-Shift-U", _("Update entire library")],
                 ["Menu", _("Display popup menu")],
                 ["Escape", _("Minimize to system tray (if enabled)")]]
        playbackshortcuts = \
                [["Ctrl-Left", _("Previous track")],
                 ["Ctrl-Right", _("Next track")],
                 ["Ctrl-P", _("Play/Pause")],
                 ["Ctrl-S", _("Stop")],
                 ["Ctrl-Minus", _("Lower the volume")],
                 ["Ctrl-Plus", _("Raise the volume")]]
        currentshortcuts = \
                [["Enter/Space", _("Play selected song")],
                 ["Delete", _("Remove selected song(s)")],
                 ["Ctrl-I", _("Center currently playing song")],
                 ["Ctrl-T", _("Edit selected song's tags")],
                 ["Ctrl-Shift-S", _("Save to new playlist")],
                 ["Ctrl-Delete", _("Clear list")],
                 ["Alt-R", _("Randomize list")]]
        libraryshortcuts = \
                [["Enter/Space", _("Add selected song(s) or enter directory")],
                 ["Backspace", _("Go to parent directory")],
                 ["Ctrl-D", _("Add selected item(s)")],
                 ["Ctrl-R", _("Replace with selected item(s)")],
                 ["Ctrl-T", _("Edit selected song's tags")],
                 ["Ctrl-Shift-D", _("Add selected item(s) and play")],
                 ["Ctrl-Shift-R", _("Replace with selected item(s) and play")],
                 ["Ctrl-U", _("Update selected item(s)/path(s)")]]
        playlistshortcuts = \
                [["Enter/Space", _("Add selected playlist(s)")],
                 ["Delete", _("Remove selected playlist(s)")],
                 ["Ctrl-D", _("Add selected playlist(s)")],
                 ["Ctrl-R", _("Replace with selected playlist(s)")],
                 ["Ctrl-Shift-D", _("Add selected playlist(s) and play")],
                 ["Ctrl-Shift-R", _(('Replace with selected '
                                     'playlist(s) and play'))]]
        streamshortcuts = \
                [["Enter/Space", _("Add selected stream(s)")],
                 ["Delete", _("Remove selected stream(s)")],
                 ["Ctrl-D", _("Add selected stream(s)")],
                 ["Ctrl-R", _("Replace with selected stream(s)")],
                 ["Ctrl-Shift-D", _("Add selected stream(s) and play")],
                 ["Ctrl-Shift-R", _(('Replace with selected '
                                     'stream(s) and play'))]]
        infoshortcuts = \
                [["Ctrl-T", _("Edit playing song's tags")]]
        # define the main array- this adds headings to each section of
        # shortcuts that will be displayed
        shortcuts = [[_("Main Shortcuts"), mainshortcuts],
                [_("Playback Shortcuts"), playbackshortcuts],
                [_("Current Shortcuts"), currentshortcuts],
                [_("Library Shortcuts"), libraryshortcuts],
                [_("Playlist Shortcuts"), playlistshortcuts],
                [_("Stream Shortcuts"), streamshortcuts],
                [_("Info Shortcuts"), infoshortcuts]]
        self.shortcuts_dialog = self.builder.get_object('shortcuts_dialog')
        self.shortcuts_dialog.connect('response', self.shortcuts_close)
        self.shortcuts_dialog.connect('delete_event', self.shortcuts_close)
        shortcuts_close_button = self.builder.get_object(
            'shortcuts_dialog_closebutton')
        shortcuts_close_button.connect('clicked', self.shortcuts_close_click_cb)

        # each pair is a [ heading, shortcutlist ]
        vbox = self.builder.get_object('shortcuts_dialog_content_box')
        for heading, shortcutlist in shortcuts:
            titlelabel = Gtk.Label(heading, xalign=0)
            titlelabel.get_style_context().add_class('heading')
            vbox.pack_start(titlelabel, False, False, 2)

            # print the items of [ shortcut, desc ]
            for shortcut, desc in shortcutlist:
                tmphbox = Gtk.HBox()

                tmplabel = Gtk.Label('{}:'.format(shortcut), xalign=0)
                tmplabel.get_style_context().add_class('shortcut')
                tmpdesc = Gtk.Label(desc, xalign=0, wrap=False)

                tmphbox.pack_start(tmplabel, False, False, 2)
                tmphbox.pack_start(tmpdesc, True, True, 2)

                vbox.pack_start(tmphbox, False, False, 2)
            vbox.pack_start(Gtk.Label(" "), False, False, 2)

    def about_shortcuts(self, _button):
        if not self.shortcuts_dialog:
            self._init_shortcuts_dialog()
        self.shortcuts_dialog.show_all()
        self.shortcuts_dialog.run()

    def statstext(self, stats):
        song_count = int(stats['songs'])
        song_text = ngettext('{count} song.', '{count} songs.',
                             song_count).format(count=song_count)
        album_count = int(stats['albums'])
        album_text = ngettext('{count} album.', '{count} albums.',
                              album_count).format(count=album_count)
        artist_count = int(stats['artists'])
        artist_text = ngettext('{count} artist.', '{count} artists.',
                               artist_count).format(count=artist_count)

        try:
            db_playtime = float(stats['db_playtime'])
            hours = int(misc.convert_time(db_playtime).split(':')[-3])
        except:
            hours = 0
        if hours >= 24:
            days = round(hours / 24, 1)
            time_text = ngettext('{count} day of bliss.',
                                 '{count} days of bliss.',
                                 days).format(count=days)
        else:
            time_text = ngettext('{count} hour of bliss.',
                                 '{count} hours of bliss.',
                                 hours).format(count=hours)

        parts = (song_text, album_text, artist_text, time_text)
        live_parts = [part for part in parts if part is not None]
        return '\n'.join(live_parts)

    def about_load(self, stats):
        self.builder = ui.builder('about')
        self.provider = ui.css_provider('about')
        self.about_dialog = self.builder.get_object('about_dialog')
        try:
            self.about_dialog.set_transient_for(self.parent_window)
        except:
            pass
        self.about_dialog.set_version(self.version)
        if stats:
            self.about_dialog.set_copyright(self.statstext(stats))
        context = self.about_dialog.get_style_context()
        logo_icon = Gtk.IconFactory.lookup_default('sonata-large')
        logo_pb = logo_icon.render_icon_pixbuf(context, -1)
        self.about_dialog.set_logo(logo_pb)
        # Add button to show keybindings:
        children = self.about_dialog.action_area.get_children()[-1]
        self.about_dialog.action_area.reorder_child(children, -2)
        # Connect to callbacks
        self.about_dialog.connect('response', self.about_close)
        self.about_dialog.connect('delete_event', self.about_close)
        shortcut_button = self.builder.get_object('shortcut_button')
        shortcut_button.connect('clicked', self.about_shortcuts)
        self.about_dialog.show_all()




# Copyright 2006-2009 Scott Horowitz <stonecrest@gmail.com>
# Copyright 2009-2014 Jonathan Ballet <jon@multani.info>
#
# This file is part of Sonata.
#
# Sonata is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Sonata is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Sonata.  If not, see <http://www.gnu.org/licenses/>.

import os
import subprocess
import re
import locale
import logging
import sys

from gi.repository import GLib


logger = logging.getLogger(__name__)


def convert_time(seconds):
    """
    Converts time in seconds to 'hh:mm:ss' format
    with leading zeros as appropriate and optional hours
    """
    hours, minutes, seconds = convert_time_raw(seconds)
    if hours == 0:
       return "%02d:%02d" %(minutes, seconds)
    return "%02d:%02d:%02d" %(hours, minutes, seconds)

def convert_time_raw(seconds):
    hours = seconds // 3600
    seconds -= 3600 * hours
    minutes = seconds // 60
    seconds -= 60 * minutes
    return hours, minutes, seconds

def escape_html(s):
    if not s: # None or ""
        return ""

    # & needs to be escaped first, before more are introduced:
    s = s.replace('&', '&amp;')
    s = s.replace('<', '&lt;')
    s = s.replace('>', '&gt;')
    s = s.replace('"', '&quot;')
    return s


def unescape_html(s):
    s = s.replace('&lt;', '<')
    s = s.replace('&gt;', '>')
    s = s.replace('&quot;', '"')
    s = s.replace('&nbsp;', ' ')
    # & needs to be unescaped last, so it can't get unescaped twice
    s = s.replace('&amp;', '&')
    # FIXME why did we have this too? s = s.replace('amp;', '&')
    return s


def wiki_to_html(s):
    # XXX Should we depend on a library to do this or get
    # html from the services?
    s = re.sub(r"'''''(.*?)'''''", r"<i><b>\1</b></i>", s)
    s = re.sub(r"'''(.*?)'''", r"<b>\1</b>", s)
    s = re.sub(r"''(.*?)''", r"<i>\1</i>", s)
    return s


def strip_all_slashes(s):
    if not s: # None or ""
        return ""
    s = s.replace("\\", "")
    s = s.replace("/", "")
    s = s.replace("\"", "")
    return s


def _rmgeneric(path, __func__):
    try:
        __func__(path)
    except OSError:
        pass


def link_markup(s, enclose_in_parentheses, small, linkcolor):
    if enclose_in_parentheses:
        s = "(%s)" % s
    if small:
        s = "<small>%s</small>" % s
    if linkcolor:
        color = linkcolor
    else:
        color = "blue" #no theme color, default to blue..
    s = "<span color='%s'>%s</span>" % (color, s)
    return s


def iunique(iterable, key=id):
    seen = set()
    for i in iterable:
        if key(i) not in seen:
            seen.add(key(i))
            yield i


def remove_list_duplicates(inputlist, case=True):
    # Note that we can't use list(set(inputlist))
    # because we want the inputlist order preserved.
    if case:
        key = lambda x: x
    else:
        # repr() allows inputlist to be a list of tuples
        # FIXME: Doesn't correctly compare uppercase and
        # lowercase unicode
        key = lambda x: repr(x).lower()
    return list(iunique(inputlist, key))

the_re = re.compile('^the ')


def lower_no_the(s):
    s = the_re.sub('', s.lower())
    s = str(s)
    return s


def create_dir(dirname):
    if not os.path.exists(os.path.expanduser(dirname)):
        try:
            os.makedirs(os.path.expanduser(dirname))
        except (IOError, OSError):
            pass


def remove_file(filename):
    if os.path.exists(filename):
        try:
            os.remove(filename)
        except:
            pass


def remove_dir_recursive(path):
    if not os.path.isdir(path):
        return

    files = os.listdir(path)

    for x in files:
        fullpath = os.path.join(path, x)
        if os.path.isfile(fullpath):
            f = os.remove
            _rmgeneric(fullpath, f)
        elif os.path.isdir(fullpath):
            remove_dir_recursive(fullpath)
            f = os.rmdir
            _rmgeneric(fullpath, f)


def file_exists_insensitive(filename):
    # Returns an updated filename that exists on the
    # user's filesystem; checks all possible combinations
    # of case.
    if os.path.exists(filename):
        return filename

    regexp = re.compile(re.escape(filename), re.IGNORECASE)

    path = os.path.dirname(filename)

    try:
        files = os.listdir(path)
    except OSError:
        return filename

    for x in files:
        fullpath = os.path.join(path, x)
        if regexp.match(fullpath):
            return fullpath

    return filename


def browser_load(docslink, browser, window):
    logger.debug("Loading URL: %s", docslink)
    if browser and browser.strip():
        browsers = [browser.strip()]
    else:
        browsers = ["xdg-open",    # default, this respect the used DE
                "x-www-browser", # default on Debian-based systems
                "gnome-open",
                "kde-open",
                "exo-open",
                "firefox",
                "opera",
                "chromium"]
    for browser in browsers:
        try:
            subprocess.Popen(browser.split() + [docslink])
            break # done
        except OSError:
            pass  # try next
    else: # none worked
        return False
    return True


def is_lang_rtl(window):
    from gi.repository import Pango
    # Check if a RTL (right-to-left) language:
    return window.get_pango_context().get_base_dir() == Pango.Direction.RTL


def sanitize_musicdir(mdir):
    return os.path.expanduser(mdir) if mdir else ''


def mpd_env_vars():
    host = None
    port = None
    password = None
    if 'MPD_HOST' in os.environ:
        if '@' in os.environ['MPD_HOST']:
            password, host = os.environ['MPD_HOST'].split('@')
        else:
            host = os.environ['MPD_HOST']
    if 'MPD_PORT' in os.environ:
        port = int(os.environ['MPD_PORT'])
    return (host, port, password)


def get_files_recursively(dirname):
    filenames = []
    os.walk(dirname, _get_files_recursively, filenames)
    return filenames


def _get_files_recursively(filenames, dirname, files):
    filenames.extend([os.path.join(dirname, f) for f in files])



# Copyright 2006-2009 Scott Horowitz <stonecrest@gmail.com>
# Copyright 2009-2014 Jonathan Ballet <jon@multani.info>
#
# This file is part of Sonata.
#
# Sonata is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Sonata is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Sonata.  If not, see <http://www.gnu.org/licenses/>.

import os
import re
import gettext
import locale
import threading # libsearchfilter_toggle starts thread libsearchfilter_loop
import operator

from gi.repository import Gtk, Gdk, GdkPixbuf, GObject, GLib, Pango

from sonata import ui, misc, consts, formatting, breadcrumbs, mpdhelper as mpdh
from sonata.artwork import get_multicd_album_root_dir
from sonata.song import SongRecord


VARIOUS_ARTISTS = _("Various Artists")


def list_mark_various_artists_albums(albums):
    for i in range(len(albums)):
        if i + consts.NUM_ARTISTS_FOR_VA - 1 > len(albums)-1:
            break
        VA = False
        for j in range(1, consts.NUM_ARTISTS_FOR_VA):
            if albums[i].album.lower() != albums[i + j].album.lower() or \
               albums[i].year  != albums[i + j].year or \
               albums[i].path  != albums[i + j].path:
                break
            if albums[i].artist == albums[i + j].artist:
                albums.pop(i + j)
                break
            if j == consts.NUM_ARTISTS_FOR_VA - 1:
                VA = True
        if VA:
            album = albums[i]._asdict()
            album['artist'] = VARIOUS_ARTISTS
            albums[i] = SongRecord(**album)
            j = 1
            while i + j <= len(albums) - 1:
                if albums[i].album.lower() == albums[i + j].album.lower() \
                   and albums[i].year == albums[i + j].year:
                    albums.pop(i + j)
                else:
                    break
    return albums


class Library:
    def __init__(self, config, mpd, artwork, TAB_LIBRARY, settings_save,
                 filter_key_pressed, on_add_item, connected,
                 on_library_button_press, add_tab):
        self.artwork = artwork
        self.config = config
        self.mpd = mpd
        self.librarymenu = None # cyclic dependency, set later
        self.settings_save = settings_save
        self.filter_key_pressed = filter_key_pressed
        self.on_add_item = on_add_item
        self.connected = connected
        self.on_library_button_press = on_library_button_press

        self.NOTAG = _("Untagged")
        self.search_terms = [_('Artist'), _('Title'), _('Album'), _('Genre'),
                             _('Filename'), _('Everything')]
        self.search_terms_mpd = ['artist', 'title', 'album', 'genre', 'file',
                                 'any']

        self.libfilterbox_cmd_buf = None
        self.libfilterbox_cond = None
        self.libfilterbox_source = None

        self.prevlibtodo_base = None
        self.prevlibtodo_base_results = None
        self.prevlibtodo = None

        self.save_timeout = None
        self.libsearch_last_tooltip = None

        self.lib_view_filesystem_cache = None
        self.lib_view_artist_cache = None
        self.lib_view_genre_cache = None
        self.lib_view_album_cache = None
        self.lib_list_genres = None
        self.lib_list_artists = None
        self.lib_list_albums = None
        self.lib_list_years = None
        self.view_caches_reset()

        # Library tab
        self.builder = ui.builder('library')
        self.css_provider = ui.css_provider('library')

        self.libraryvbox = self.builder.get_object('library_page_v_box')
        self.library = self.builder.get_object('library_page_treeview')
        self.library_selection = self.library.get_selection()
        self.breadcrumbs = self.builder.get_object('library_crumbs_box')
        self.crumb_section = self.builder.get_object(
            'library_crumb_section_togglebutton')
        self.crumb_section_image = self.builder.get_object(
            'library_crumb_section_image')
        self.crumb_break = self.builder.get_object(
            'library_crumb_break_box')
        self.breadcrumbs.set_crumb_break(self.crumb_break)
        self.crumb_section_handler = None
        expanderwindow2 = self.builder.get_object('library_page_scrolledwindow')
        self.searchbox = self.builder.get_object('library_page_searchbox')
        self.searchcombo = self.builder.get_object('library_page_searchbox_combo')
        self.searchtext = self.builder.get_object('library_page_searchbox_entry')
        self.searchbutton = self.builder.get_object('library_page_searchbox_button')
        self.searchbutton.hide()
        self.libraryview = self.builder.get_object('library_crumb_button')
        self.tab_label_widget = self.builder.get_object('library_tab_eventbox')
        tab_label = self.builder.get_object('library_tab_label')
        tab_label.set_text(TAB_LIBRARY)

        self.tab = add_tab(self.libraryvbox, self.tab_label_widget,
                           TAB_LIBRARY, self.library)

        # Assign some pixbufs for use in self.library
        self.openpb2 = self.library.render_icon(Gtk.STOCK_OPEN,
                                                Gtk.IconSize.LARGE_TOOLBAR)
        self.harddiskpb2 = self.library.render_icon(Gtk.STOCK_HARDDISK,
                                                   Gtk.IconSize.LARGE_TOOLBAR)
        self.openpb = self.library.render_icon(Gtk.STOCK_OPEN,
                                               Gtk.IconSize.MENU)
        self.harddiskpb = self.library.render_icon(Gtk.STOCK_HARDDISK,
                                                   Gtk.IconSize.MENU)
        self.albumpb = self.library.render_icon('sonata-album',
                                                Gtk.IconSize.LARGE_TOOLBAR)
        self.genrepb = self.library.render_icon('gtk-orientation-portrait',
                                                Gtk.IconSize.LARGE_TOOLBAR)
        self.artistpb = self.library.render_icon('sonata-artist',
                                                 Gtk.IconSize.LARGE_TOOLBAR)
        self.sonatapb = self.library.render_icon('sonata',
                                                 Gtk.IconSize.LARGE_TOOLBAR)

        # list of the library views: (id, name, icon name, label)
        self.VIEWS = [
            (consts.VIEW_FILESYSTEM, 'filesystem',
             Gtk.STOCK_HARDDISK, _("Filesystem")),
            (consts.VIEW_ALBUM, 'album',
             'sonata-album', _("Albums")),
            (consts.VIEW_ARTIST, 'artist',
             'sonata-artist', _("Artists")),
            (consts.VIEW_GENRE, 'genre',
             Gtk.STOCK_ORIENTATION_PORTRAIT, _("Genres")),
            ]

        self.library.connect('row_activated', self.on_library_row_activated)
        self.library.connect('button_press_event',
                             self.on_library_button_press)
        self.library.connect('key-press-event', self.on_library_key_press)
        self.library.connect('query-tooltip', self.on_library_query_tooltip)
        expanderwindow2.connect('scroll-event', self.on_library_scrolled)
        self.libraryview.connect('clicked', self.library_view_popup)
        self.searchtext.connect('key-press-event',
                                self.libsearchfilter_key_pressed)
        self.searchtext.connect('activate', self.libsearchfilter_on_enter)
        self.searchbutton.connect('clicked', self.on_search_end)

        self.libfilter_changed_handler = self.searchtext.connect(
            'changed', self.libsearchfilter_feed_loop)
        searchcombo_changed_handler = self.searchcombo.connect(
            'changed', self.on_library_search_combo_change)

        # Initialize library data and widget
        self.libraryposition = {}
        self.libraryselectedpath = {}
        self.searchcombo.handler_block(searchcombo_changed_handler)
        self.searchcombo.set_active(self.config.last_search_num)
        self.searchcombo.handler_unblock(searchcombo_changed_handler)
        self.librarydata = Gtk.ListStore(GdkPixbuf.Pixbuf,
                                         GObject.TYPE_PYOBJECT, str)
        self.library.set_model(self.librarydata)
        self.library.set_search_column(2)
        self.librarycell = Gtk.CellRendererText()
        self.librarycell.set_property("ellipsize", Pango.EllipsizeMode.END)
        self.libraryimg = Gtk.CellRendererPixbuf()
        self.librarycolumn = Gtk.TreeViewColumn()
        self.librarycolumn.pack_start(self.libraryimg, False)
        self.librarycolumn.pack_start(self.librarycell, True)
        self.librarycolumn.add_attribute(self.libraryimg, 'pixbuf', 0)
        self.librarycolumn.add_attribute(self.librarycell, 'markup', 2)
        self.librarycolumn.set_sizing(Gtk.TreeViewColumnSizing.AUTOSIZE)
        self.library.append_column(self.librarycolumn)
        self.library_selection.set_mode(Gtk.SelectionMode.MULTIPLE)

    def get_libraryactions(self):
        return [(name + 'view', icon, label,
             None, None, self.on_libraryview_chosen)
            for _view, name, icon, label in self.VIEWS]

    def get_model(self):
        return self.librarydata

    def get_widgets(self):
        return self.libraryvbox

    def get_treeview(self):
        return self.library

    def get_selection(self):
        return self.library_selection

    def set_librarymenu(self, librarymenu):
        self.librarymenu = librarymenu
        self.librarymenu.attach_to_widget(self.libraryview, None)

    def library_view_popup(self, button):
        if hasattr(self.librarymenu, 'popup_at_widget'):
            self.librarymenu.popup_at_widget(button,
                                             Gdk.Gravity.CENTER,
                                             Gdk.Gravity.NORTH_WEST,
                                             None)
        else:
            self.librarymenu.popup(None, None,
                                   self.library_view_position_menu,
                                   button, 1, 0)

    def library_view_position_menu(self, _menu, *args):
        # Depending on PyGObject version, args is either just the
        # button that was our user_data, or [x, y, button].
        button = args[-1]
        alloc = button.get_allocation()
        return (self.config.x + alloc.x,
                self.config.y + alloc.y + alloc.height,
                True)

    def on_libraryview_chosen(self, action):
        if self.search_visible():
            self.on_search_end(None)
        if action.get_name() == 'filesystemview':
            self.config.lib_view = consts.VIEW_FILESYSTEM
        elif action.get_name() == 'artistview':
            self.config.lib_view = consts.VIEW_ARTIST
        elif action.get_name() == 'genreview':
            self.config.lib_view = consts.VIEW_GENRE
        elif action.get_name() == 'albumview':
            self.config.lib_view = consts.VIEW_ALBUM
        self.library.grab_focus()
        self.libraryposition = {}
        self.libraryselectedpath = {}
        self.library_browse(root=SongRecord(path="/"))
        try:
            if len(self.librarydata) > 0:
                first = Gtk.TreePath.new_first()
                to = Gtk.TreePath.new()
                to.append_index(len(self.librarydata) - 1)
                self.library_selection.unselect_range(first, to)
        except Exception as e:
            # XXX import logger here in the future
            raise e
        GLib.idle_add(self.library.scroll_to_point, 0, 0)

    def view_caches_reset(self):
        # We should call this on first load and whenever mpd is
        # updated.
        self.lib_view_filesystem_cache = None
        self.lib_view_artist_cache = None
        self.lib_view_genre_cache = None
        self.lib_view_album_cache = None
        self.lib_list_genres = None
        self.lib_list_artists = None
        self.lib_list_albums = None
        self.lib_list_years = None

    def on_library_scrolled(self, _widget, _event):
        try:
            # Use GLib.idle_add so that we can get the visible
            # state of the treeview
            GLib.idle_add(self._on_library_scrolled)
        except:
            pass

    def _on_library_scrolled(self):
        if not self.config.show_covers:
            return

        # This avoids a warning about a NULL node in get_visible_range
        if not self.library.props.visible:
            return

        visible_range = self.library.get_visible_range()

        if visible_range is None:
            return
        else:
            start_row, end_row = visible_range

        self.artwork.library_artwork_update(self.librarydata, start_row,
                                            end_row, self.albumpb)

    def library_browse(self, _widget=None, root=None):
        # Populates the library list with entries
        if not self.connected():
            return

        if root is None or (self.config.lib_view == consts.VIEW_FILESYSTEM \
                            and root.path is None):
            root = SongRecord(path="/")
        if self.config.wd is None or (self.config.lib_view == \
                                      consts.VIEW_FILESYSTEM and \
                                      self.config.wd.path is None):
            self.config.wd = SongRecord(path="/")

        prev_selection = []
        prev_selection_root = False
        prev_selection_parent = False
        if root == self.config.wd:
            # This will happen when the database is updated. So, lets save
            # the current selection in order to try to re-select it after
            # the update is over.
            model, selected = self.library_selection.get_selected_rows()
            for path in selected:
                prev_selection.append(model.get_value(model.get_iter(path), 1))
            self.libraryposition[self.config.wd] = \
                    self.library.get_visible_rect().width
            path_updated = True
        else:
            path_updated = False

        new_level = self.library_get_data_level(root)
        curr_level = self.library_get_data_level(self.config.wd)
        # The logic below is more consistent with, e.g., thunar.
        if new_level > curr_level:
            # Save position and row for where we just were if we've
            # navigated into a sub-directory:
            self.libraryposition[self.config.wd] = \
                    self.library.get_visible_rect().width
            model, rows = self.library_selection.get_selected_rows()
            if len(rows) > 0:
                data = self.librarydata.get_value(
                    self.librarydata.get_iter(rows[0]), 2)
                self.libraryselectedpath[self.config.wd] = rows[0]
        elif (self.config.lib_view == consts.VIEW_FILESYSTEM and \
              root != self.config.wd) \
        or (self.config.lib_view != consts.VIEW_FILESYSTEM and new_level != \
            curr_level):
            # If we've navigated to a parent directory, don't save
            # anything so that the user will enter that subdirectory
            # again at the top position with nothing selected
            self.libraryposition[self.config.wd] = 0
            self.libraryselectedpath[self.config.wd] = None

        # In case sonata is killed or crashes, we'll save the library state
        # in 5 seconds (first removing any current settings_save timeouts)
        if self.config.wd != root:
            try:
                GLib.source_remove(self.save_timeout)
            except:
                pass
            self.save_timeout = GLib.timeout_add(5000, self.settings_save)

        self.config.wd = root
        self.library.freeze_child_notify()
        self.library.set_model(None)
        self.librarydata.clear()

        # Populate treeview with data:
        bd = []
        wd = self.config.wd
        while len(bd) == 0:
            if self.config.lib_view == consts.VIEW_FILESYSTEM:
                bd = self.library_populate_filesystem_data(wd.path)
            elif self.config.lib_view == consts.VIEW_ALBUM:
                if wd.album is not None:
                    bd = self.library_populate_data(artist=wd.artist,
                                                    album=wd.album,
                                                    year=wd.year)
                else:
                    bd = self.library_populate_toplevel_data(albumview=True)
            elif self.config.lib_view == consts.VIEW_ARTIST:
                if wd.artist is not None and wd.album is not None:
                    bd = self.library_populate_data(artist=wd.artist,
                                                    album=wd.album,
                                                    year=wd.year)
                elif self.config.wd.artist is not None:
                    bd = self.library_populate_data(artist=wd.artist)
                else:
                    bd = self.library_populate_toplevel_data(artistview=True)
            elif self.config.lib_view == consts.VIEW_GENRE:
                if wd.genre is not None and \
                   wd.artist is not None and \
                   wd.album is not None:
                    bd = self.library_populate_data(genre=wd.genre,
                                                    artist=wd.artist,
                                                    album=wd.album,
                                                    year=wd.year)
                elif wd.genre is not None:
                    bd = self.library_populate_data(genre=wd.genre,
                                                    artist=wd.artist)
                else:
                    bd = self.library_populate_toplevel_data(genreview=True)

            if len(bd) == 0:
                # Nothing found; go up a level until we reach the top level
                # or results are found
                last_wd = self.config.wd
                self.config.wd = self.library_get_parent()
                if self.config.wd == last_wd:
                    break

        for _sort, path in bd:
            self.librarydata.append(path)

        self.library.set_model(self.librarydata)
        self.library.thaw_child_notify()

        # Scroll back to set view for current dir:
        self.library.realize()
        GLib.idle_add(self.library_set_view, not path_updated)
        if len(prev_selection) > 0 or prev_selection_root or \
           prev_selection_parent:
            # Retain pre-update selection:
            self.library_retain_selection(prev_selection, prev_selection_root,
                                          prev_selection_parent)

        self.update_breadcrumbs()

    def update_breadcrumbs(self):
        # remove previous buttons
        for b in self.breadcrumbs:
            self.breadcrumbs.remove(b)

        # find info for current view
        view, _name, icon, label = [v for v in self.VIEWS
                          if v[0] == self.config.lib_view][0]

        # the first crumb is the root of the current view
        self.crumb_section.set_label(label)
        self.crumb_section_image.set_from_stock(icon, Gtk.IconSize.MENU)
        self.crumb_section.set_tooltip_text(label)
        if self.crumb_section_handler:
            self.crumb_section.disconnect(self.crumb_section_handler)


        crumbs = []
        # crumbs are specific to the view
        if view == consts.VIEW_FILESYSTEM:
            if self.config.wd.path and self.config.wd.path != '/':
                parts = self.config.wd.path.split('/')
            else:
                parts = [] # no crumbs for /
            # append a crumb for each part
            for i, part in enumerate(parts):
                partpath = '/'.join(parts[:i + 1])
                target = SongRecord(path=partpath)
                crumbs.append((part, Gtk.STOCK_OPEN, None, target))
        else:
            parts = ()
            if view == consts.VIEW_ALBUM:
                # We don't want to show an artist button in album view
                keys = 'genre', 'album'
                nkeys = 2
                parts = (self.config.wd.genre, self.config.wd.album)
            else:
                keys = 'genre', 'artist', 'album'
                nkeys = 3
                parts = (self.config.wd.genre, self.config.wd.artist,
                         self.config.wd.album)
            # append a crumb for each part
            for i, key, part in zip(range(nkeys), keys, parts):
                if part is None:
                    continue
                partdata = dict(list(zip(keys, parts))[:i + 1])
                target = SongRecord(**partdata)
                pb, icon = None, None
                if key == 'album':
                    # Album artwork, with self.alumbpb as a backup:
                    cache_data = SongRecord(artist=self.config.wd.artist,
                                            album=self.config.wd.album,
                                            path=self.config.wd.path)
                    pb = self.artwork.cache.get_pixbuf(cache_data,
                                                       consts.LIB_COVER_SIZE)
                    if pb is None:
                        icon = 'album'
                elif key == 'artist':
                    icon = 'sonata-artist'
                else:
                    icon = Gtk.STOCK_ORIENTATION_PORTRAIT
                crumbs.append((part, icon, pb, target))

        if not len(crumbs):
            self.crumb_section.set_active(True)
            context = self.crumb_section.get_style_context()
            context.add_class('last_crumb')
        else:
            self.crumb_section.set_active(False)
            context = self.crumb_section.get_style_context()
            context.remove_class('last_crumb')

        self.crumb_section_handler = self.crumb_section.connect('toggled',
            self.library_browse, SongRecord(path='/'))

        # add a button for each crumb
        for crumb in crumbs:
            text, icon, pb, target = crumb
            text = misc.escape_html(text)
            label = Gtk.Label(text, use_markup=True)

            if icon:
                image = Gtk.Image.new_from_stock(icon, Gtk.IconSize.MENU)
            elif pb:
                pb = pb.scale_simple(16, 16, GdkPixbuf.InterpType.HYPER)
                image = Gtk.Image.new_from_pixbuf(pb)

            b = breadcrumbs.CrumbButton(image, label)

            if crumb is crumbs[-1]:
                # FIXME makes the button request minimal space:
                b.set_active(True)
                context = b.get_style_context()
                context.add_class('last_crumb')

            b.set_tooltip_text(label.get_label())
            b.connect('toggled', self.library_browse, target)
            self.breadcrumbs.pack_start(b, False, False, 0)
            b.show_all()

    def library_populate_filesystem_data(self, path):
        # List all dirs/files at path
        bd = []
        if path == '/' and self.lib_view_filesystem_cache is not None:
            # Use cache if possible...
            bd = self.lib_view_filesystem_cache
        else:
            for item in self.mpd.lsinfo(path):
                if 'directory' in item:
                    name = os.path.basename(item['directory'])
                    data = SongRecord(path=item["directory"])
                    bd += [('d' + str(name).lower(), [self.openpb, data,
                                                      misc.escape_html(name)])]
                elif 'file' in item:
                    data = SongRecord(path=item['file'])
                    bd += [('f' + item['file'].lower(),
                            [self.sonatapb, data,
                             formatting.parse(self.config.libraryformat, item,
                                              True)])]
            bd.sort(key=operator.itemgetter(0))
        return bd

    def library_get_toplevel_cache(self, genreview=False, artistview=False,
                                   albumview=False):
        if genreview and self.lib_view_genre_cache is not None:
            bd = self.lib_view_genre_cache
        elif artistview and self.lib_view_artist_cache is not None:
            bd = self.lib_view_artist_cache
        elif albumview and self.lib_view_album_cache is not None:
            bd = self.lib_view_album_cache
        else:
            return None
        # Check if we can update any artwork:
        for _sort, info in bd:
            pb = info[0]
            if pb == self.albumpb:
                key = SongRecord(path=info[1].path, artist=info[1].artist,
                                 album=info[1].album)
                pb2 = self.artwork.cache.get_pixbuf(key,
                                                    consts.LIB_COVER_SIZE)
                if pb2 is not None:
                    info[0] = pb2
        return bd

    def library_populate_toplevel_data(self, genreview=False, artistview=False,
                                       albumview=False):
        bd = self.library_get_toplevel_cache(genreview, artistview, albumview)
        if bd is not None:
            # We have our cached data, woot.
            return bd
        bd = []
        if genreview or artistview:
            # Only for artist/genre views, album view is handled differently
            # since multiple artists can have the same album name
            if genreview:
                items = self.library_return_list_items('genre')
                pb = self.genrepb
            else:
                items = self.library_return_list_items('artist')
                pb = self.artistpb
            if not (self.NOTAG in items):
                items.append(self.NOTAG)
            for item in items:
                if genreview:
                    playtime, num_songs = self.library_return_count(genre=item)
                    data = SongRecord(genre=item)
                else:
                    playtime, num_songs = self.library_return_count(
                        artist=item)
                    data = SongRecord(artist=item)
                if num_songs > 0:
                    display = misc.escape_html(item)
                    display += self.add_display_info(num_songs, playtime)
                    bd += [(misc.lower_no_the(item), [pb, data, display])]
        elif albumview:
            albums = []
            untagged_found = False
            for item in self.mpd.listallinfo('/'):
                if 'file' in item and 'album' in item:
                    album = item['album']
                    artist = item.get('artist', self.NOTAG)
                    year = item.get('date', self.NOTAG)
                    path = get_multicd_album_root_dir(
                        os.path.dirname(item['file']))
                    data = SongRecord(album=album, artist=artist,
                                      year=year, path=path)
                    albums.append(data)
                    if album == self.NOTAG:
                        untagged_found = True
            if not untagged_found:
                albums.append(SongRecord(album=self.NOTAG))
            albums = misc.remove_list_duplicates(albums, case=False)
            albums = list_mark_various_artists_albums(albums)
            for item in albums:
                album, artist, _genre, year, path = item
                playtime, num_songs = self.library_return_count(artist=artist,
                                                                album=album,
                                                                year=year)
                if num_songs > 0:
                    data = SongRecord(artist=artist, album=album,
                                           year=year, path=path)
                    display = misc.escape_html(album)
                    if artist and year and len(artist) > 0 and len(year) > 0 \
                       and artist != self.NOTAG and year != self.NOTAG:
                        display += " <span weight='light'>(%s, %s)</span>" \
                                % (misc.escape_html(artist),
                                   misc.escape_html(year))
                    elif artist and len(artist) > 0 and artist != self.NOTAG:
                        display += " <span weight='light'>(%s)</span>" \
                                % misc.escape_html(artist)
                    elif year and len(year) > 0 and year != self.NOTAG:
                        display += " <span weight='light'>(%s)</span>" \
                                % misc.escape_html(year)
                    display += self.add_display_info(num_songs, playtime)
                    bd += [(misc.lower_no_the(album), [self.albumpb, data,
                                                       display])]
        bd.sort(key=lambda key: locale.strxfrm(key[0]))
        if genreview:
            self.lib_view_genre_cache = bd
        elif artistview:
            self.lib_view_artist_cache = bd
        elif albumview:
            self.lib_view_album_cache = bd
        return bd


    def library_populate_data(self, genre=None, artist=None, album=None,
                              year=None):
        # Create treeview model info
        bd = []
        if genre is not None and artist is None and album is None:
            # Artists within a genre
            artists = self.library_return_list_items('artist', genre=genre)
            if len(artists) > 0:
                if not self.NOTAG in artists:
                    artists.append(self.NOTAG)
                for artist in artists:
                    playtime, num_songs = self.library_return_count(
                        genre=genre, artist=artist)
                    if num_songs > 0:
                        display = misc.escape_html(artist)
                        display += self.add_display_info(num_songs, playtime)
                        data = SongRecord(genre=genre, artist=artist)
                        bd += [(misc.lower_no_the(artist),
                                [self.artistpb, data, display])]
        elif artist is not None and album is None:
            # Albums/songs within an artist and possibly genre
            # Albums first:
            if genre is not None:
                albums = self.library_return_list_items('album', genre=genre,
                                                        artist=artist)
            else:
                albums = self.library_return_list_items('album', artist=artist)
            for album in albums:
                if genre is not None:
                    years = self.library_return_list_items('date', genre=genre,
                                                           artist=artist,
                                                           album=album)
                else:
                    years = self.library_return_list_items('date',
                                                           artist=artist,
                                                           album=album)
                if not self.NOTAG in years:
                    years.append(self.NOTAG)
                for year in years:
                    if genre is not None:
                        playtime, num_songs = self.library_return_count(
                            genre=genre, artist=artist, album=album, year=year)
                        if num_songs > 0:
                            files = self.library_return_list_items(
                                'file', genre=genre, artist=artist,
                                album=album, year=year)
                            path = os.path.dirname(files[0])
                            data = SongRecord(genre=genre, artist=artist,
                                              album=album, year=year, path=path)
                    else:
                        playtime, num_songs = self.library_return_count(
                            artist=artist, album=album, year=year)
                        if num_songs > 0:
                            files = self.library_return_list_items(
                                'file', artist=artist, album=album, year=year)
                            path = os.path.dirname(files[0])
                        cache_data = SongRecord(artist=artist, album=album,
                                                path=path)
                        data = SongRecord(artist=artist, album=album,
                                          year=year, path=path)
                    if num_songs > 0:
                        cache_data = SongRecord(artist=artist, album=album, path=path)
                        display = misc.escape_html(album)
                        if year and len(year) > 0 and year != self.NOTAG:
                            display += " <span weight='light'>(%s)</span>" \
                                    % misc.escape_html(year)
                        display += self.add_display_info(num_songs, playtime)
                        ordered_year = year
                        if ordered_year == self.NOTAG:
                            ordered_year = '9999'
                        pb = self.artwork.cache.get_pixbuf(
                            cache_data, consts.LIB_COVER_SIZE,
                            self.albumpb)
                        bd += [(ordered_year + misc.lower_no_the(album),
                                [pb, data, display])]
            # Now, songs not in albums:
            bd += self.library_populate_data_songs(genre, artist, self.NOTAG,
                                                   None)
        else:
            # Songs within an album, artist, year, and possibly genre
            bd += self.library_populate_data_songs(genre, artist, album, year)
        bd.sort(key=lambda key: locale.strxfrm(key[0]))
        return bd

    def library_populate_data_songs(self, genre, artist, album, year):
        bd = []
        if genre is not None:
            songs, _playtime, _num_songs = \
            self.library_return_search_items(genre=genre, artist=artist,
                                             album=album, year=year)
        else:
            songs, _playtime, _num_songs = self.library_return_search_items(
                artist=artist, album=album, year=year)
        for song in songs:
            data = SongRecord(path=song.file)
            track = str(song.get('track', 99)).zfill(2)
            disc = str(song.get('disc', 99)).zfill(2)
            try:
                bd += [('f' + disc + track + misc.lower_no_the(song.title),
                        [self.sonatapb, data, formatting.parse(
                            self.config.libraryformat, song, True)])]
            except:
                bd += [('f' + disc + track + song.file.lower(),
                        [self.sonatapb, data,
                         formatting.parse(self.config.libraryformat, song,
                                          True)])]
        return bd

    def library_return_list_items(self, itemtype, genre=None, artist=None,
                                  album=None, year=None, ignore_case=True):
        # Returns all items of tag 'itemtype', in alphabetical order,
        # using mpd's 'list'. If searchtype is passed, use
        # a case insensitive search, via additional 'list'
        # queries, since using a single 'list' call will be
        # case sensitive.
        results = []
        searches = self.library_compose_list_count_searchlist(genre, artist,
                                                              album, year)
        if len(searches) > 0:
            for s in searches:
                # If we have untagged tags (''), use search instead
                # of list because list will not return anything.
                if '' in s:
                    items = []
                    songs, playtime, num_songs = \
                            self.library_return_search_items(genre, artist,
                                                             album, year)
                    for song in songs:
                        items.append(song.get(itemtype))
                else:
                    items = self.mpd.list(itemtype, *s)
                for item in items:
                    if isinstance(item, dict):
                        if itemtype in item:
                            item = item[itemtype]
                        else:
                            continue
                    if len(item) > 0:
                        results.append(item)
        else:
            if genre is None and artist is None and album is None and year \
               is None:
                for item in self.mpd.list(itemtype):
                    if isinstance(item, dict):
                        if itemtype in item:
                            item = item[itemtype]
                        else:
                            continue
                    if len(item) > 0:
                        results.append(item)
        if ignore_case:
            results = misc.remove_list_duplicates(results, case=False)
        results.sort(key=locale.strxfrm)
        return results

    def library_return_count(self, genre=None, artist=None, album=None,
                             year=None):
        # Because mpd's 'count' is case sensitive, we have to
        # determine all equivalent items (case insensitive) and
        # call 'count' for each of them. Using 'list' + 'count'
        # involves much less data to be transferred back and
        # forth than to use 'search' and count manually.
        searches = self.library_compose_list_count_searchlist(genre, artist,
                                                              album, year)
        playtime = 0
        num_songs = 0
        for s in searches:
            count = self.mpd.count(*s)
            playtime += count.playtime
            num_songs += count.songs

        return (playtime, num_songs)

    def library_compose_list_count_searchlist_single(self, search, typename,
                                                     cached_list, searchlist):
        s = []
        skip_type = (typename == 'artist' and search == VARIOUS_ARTISTS)
        if search is not None and not skip_type:
            if search == self.NOTAG:
                itemlist = [search, '']
            else:
                itemlist = []
                if cached_list is None:
                    cached_list = self.library_return_list_items(typename,
                                                             ignore_case=False)
                    # This allows us to match untagged items
                    cached_list.append('')
                for item in cached_list:
                    if str(item).lower() == str(search).lower():
                        itemlist.append(item)
            if len(itemlist) == 0:
                # There should be no results!
                return None, cached_list
            for item in itemlist:
                if len(searchlist) > 0:
                    for item2 in searchlist:
                        s.append(item2 + (typename, item))
                else:
                    s.append((typename, item))
        else:
            s = searchlist
        return s, cached_list

    def library_compose_list_count_searchlist(self, genre=None, artist=None,
                                              album=None, year=None):
        s = []
        s, self.lib_list_genres = \
                self.library_compose_list_count_searchlist_single(
                    genre, 'genre', self.lib_list_genres, s)
        if s is None:
            return []
        s, self.lib_list_artists = \
                self.library_compose_list_count_searchlist_single(
                    artist, 'artist', self.lib_list_artists, s)
        if s is None:
            return []
        s, self.lib_list_albums = \
                self.library_compose_list_count_searchlist_single(
                    album, 'album', self.lib_list_albums, s)
        if s is None:
            return []
        s, self.lib_list_years = \
                self.library_compose_list_count_searchlist_single(
                    year, 'date', self.lib_list_years, s)
        if s is None:
            return []
        return s

    def library_compose_search_searchlist_single(self, search, typename,
                                                 searchlist):
        s = []
        skip_type = (typename == 'artist' and search == VARIOUS_ARTISTS)
        if search is not None and not skip_type:
            if search == self.NOTAG:
                itemlist = [search, '']
            else:
                itemlist = [search]
            for item in itemlist:
                if len(searchlist) > 0:
                    for item2 in searchlist:
                        s.append(item2 + (typename, item))
                else:
                    s.append((typename, item))
        else:
            s = searchlist
        return s

    def library_compose_search_searchlist(self, genre=None, artist=None,
                                          album=None, year=None):
        s = []
        s = self.library_compose_search_searchlist_single(genre, 'genre', s)
        s = self.library_compose_search_searchlist_single(album, 'album', s)
        s = self.library_compose_search_searchlist_single(artist, 'artist', s)
        s = self.library_compose_search_searchlist_single(year, 'date', s)
        return s

    def library_return_search_items(self, genre=None, artist=None, album=None,
                                    year=None):
        # Returns all mpd items, using mpd's 'search', along with
        # playtime and num_songs.
        searches = self.library_compose_search_searchlist(genre, artist, album,
                                                          year)
        for s in searches:
            args_tuple = tuple(map(str, s))
            playtime = 0
            num_songs = 0
            results = []
            strip_type = None

            if len(args_tuple) == 0:
                return None, 0, 0

            items = self.mpd.search(*args_tuple)
            if items is not None:
                for item in items:
                    if strip_type is None or (strip_type is not None and not \
                                              strip_type in item.keys()):
                        match = True
                        pos = 0
                        # Ensure that if, e.g., "foo" is searched,
                        # "foobar" isn't returned too
                        for arg in args_tuple[::2]:
                            if arg in item and \
                               str(item.get(arg, '')).upper() != \
                               str(args_tuple[pos + 1]).upper():
                                match = False
                                break
                            pos += 2
                        if match:
                            results.append(item)
                            num_songs += 1
                            playtime += item.time
        return (results, int(playtime), num_songs)

    def add_display_info(self, num_songs, playtime):
        seconds = int(playtime)
        hours   = seconds // 3600
        seconds -= 3600 * hours
        minutes = seconds // 60
        seconds -= 60 * minutes
        songs_text = ngettext('{count} song', '{count} songs',
                              num_songs).format(count=num_songs)
        seconds_text = ngettext('{count} second', '{count} seconds',
                                seconds).format(count=seconds)
        minutes_text = ngettext('{count} minute', '{count} minutes',
                                minutes).format(count=minutes)
        hours_text = ngettext('{count} hour', '{count} hours',
                              hours).format(count=hours)
        time_parts = [songs_text]
        if hours > 0:
            time_parts.extend([hours_text, minutes_text])
        elif minutes > 0:
            time_parts.extend([minutes_text, seconds_text])
        else:
            time_parts.extend([seconds_text])
        display_markup = "\n<small><span weight='light'>{}</span></small>"
        display_text = ', '.join(time_parts)
        return display_markup.format(display_text)

    def library_retain_selection(self, prev_selection, prev_selection_root,
                                 prev_selection_parent):
        # Unselect everything:
        if len(self.librarydata) > 0:
            first = Gtk.TreePath.new_first()
            to = Gtk.TreePath.new()
            to.append_index(len(self.librarydata) - 1)
            self.library_selection.unselect_range(first, to)
        # Now attempt to retain the selection from before the update:
        for value in prev_selection:
            for row in self.librarydata:
                if value == row[1]:
                    self.library_selection.select_path(row.path)
                    break
        if prev_selection_root:
            self.library_selection.select_path((0,))
        if prev_selection_parent:
            self.library_selection.select_path((1,))

    def library_set_view(self, select_items=True):
        # select_items should be false if the same directory has merely
        # been refreshed (updated)
        try:
            if self.config.wd in self.libraryposition:
                self.library.scroll_to_point(
                    -1, self.libraryposition[self.config.wd])
            else:
                self.library.scroll_to_point(0, 0)
        except:
            self.library.scroll_to_point(0, 0)

        # Select and focus previously selected item
        if select_items:
            if self.config.wd in self.libraryselectedpath:
                try:
                    if self.libraryselectedpath[self.config.wd]:
                        self.library_selection.select_path(
                            self.libraryselectedpath[self.config.wd])
                        self.library.grab_focus()
                except:
                    pass

    def library_get_data_level(self, data):
        if self.config.lib_view == consts.VIEW_FILESYSTEM:
            # Returns the number of directories down:
            if data.path == '/':
                # Every other path doesn't start with "/", so
                # start the level numbering at -1
                return -1
            else:
                return data.path.count("/")
        else:
            # Returns the number of items stored in data, excluding
            # the path:
            level = 0
            for item in data:
                if item is not None:
                    level += 1
            return level

    def on_library_key_press(self, widget, event):
        if event.keyval == Gdk.keyval_from_name('Return'):
            self.on_library_row_activated(widget, widget.get_cursor()[0])
            return True

    def on_library_query_tooltip(self, widget, x, y, keyboard_mode, tooltip):
        if keyboard_mode or not self.search_visible():
            widget.set_tooltip_text("")
            return False

        bin_x, bin_y = widget.convert_widget_to_bin_window_coords(x, y)

        pathinfo = widget.get_path_at_pos(bin_x, bin_y)
        if not pathinfo:
            widget.set_tooltip_text("")
            # If the user hovers over an empty row and then back to
            # a row with a search result, this will ensure the tooltip
            # shows up again:
            GLib.idle_add(self.library_search_tooltips_enable, widget, x, y,
                          keyboard_mode, None)
            return False
        treepath, _col, _x2, _y2 = pathinfo

        i = self.librarydata.get_iter(treepath.get_indices()[0])
        path = misc.escape_html(self.librarydata.get_value(i, 1).path)
        song = self.librarydata.get_value(i, 2)
        new_tooltip = "<b>%s:</b> %s\n<b>%s:</b> %s" \
                % (_("Song"), song, _("Path"), path)

        if new_tooltip != self.libsearch_last_tooltip:
            self.libsearch_last_tooltip = new_tooltip
            self.library.set_property('has-tooltip', False)
            GLib.idle_add(self.library_search_tooltips_enable, widget, x, y,
                          keyboard_mode, tooltip)
            GLib.idle_add(widget.set_tooltip_markup, new_tooltip)
            return

        self.libsearch_last_tooltip = new_tooltip

        return False #api says we should return True, but this doesn't work?

    def library_search_tooltips_enable(self, widget, x, y, keyboard_mode,
                                       tooltip):
        self.library.set_property('has-tooltip', True)
        if tooltip is not None:
            self.on_library_query_tooltip(widget, x, y, keyboard_mode, tooltip)

    def on_library_row_activated(self, _widget, path, _column=0):
        if path is None:
            # Default to last item in selection:
            _model, selected = self.library_selection.get_selected_rows()
            if len(selected) >= 1:
                path = selected[0]
            else:
                return
        value = self.librarydata.get_value(self.librarydata.get_iter(path), 1)
        icon = self.librarydata.get_value(self.librarydata.get_iter(path), 0)
        if icon == self.sonatapb:
            # Song found, add item
            self.on_add_item(self.library)
        elif value.path == "..":
            self.library_browse_parent(None)
        else:
            self.library_browse(None, value)

    def library_get_parent(self):
        wd = self.config.wd
        if self.config.lib_view == consts.VIEW_ALBUM:
            value = SongRecord(path="/")
        elif self.config.lib_view == consts.VIEW_ARTIST:
            if wd.album is None:
                value = SongRecord(path="/")
            else:
                value = SongRecord(artist = wd.artist)
        elif self.config.lib_view == consts.VIEW_GENRE:
            if wd.album is not None:
                value = SongRecord(genre=wd.genre,
                                   artist=wd.artist)
            elif wd.artist is not None:
                value = SongRecord(genre=wd.genre)
            else:
                value = SongRecord(path="/")
        else:
            newvalue = '/'.join(wd.path.split('/')[:-1]) or '/'
            value = SongRecord(path=newvalue)
        return value

    def library_browse_parent(self, _action):
        if not self.search_visible():
            if self.library.is_focus():
                value = self.library_get_parent()
                self.library_browse(None, value)
                return True

    def not_parent_is_selected(self):
        # Returns True if something is selected and it's not
        # ".." or "/":
        model, rows = self.library_selection.get_selected_rows()
        for path in rows:
            i = model.get_iter(path)
            value = model.get_value(i, 2)
            if value != ".." and value != "/":
                return True
        return False

    def get_path_child_filenames(self, return_root, selected_only=True):
        # If return_root=True, return main directories whenever possible
        # instead of individual songs in order to reduce the number of
        # mpd calls we need to make. We won't want this behavior in some
        # instances, like when we want all end files for editing tags
        items = []
        if selected_only:
            model, rows = self.library_selection.get_selected_rows()
        else:
            model = self.librarydata
            rows = [(i,) for i in range(len(model))]
        for path in rows:
            i = model.get_iter(path)
            pb = model.get_value(i, 0)
            data = model.get_value(i, 1)
            value = model.get_value(i, 2)
            if value != ".." and value != "/":
                if data.path is not None and data.album is None and data.artist is None and \
                   data.year is None and data.genre is None:
                    if pb == self.sonatapb:
                        # File
                        items.append(data.path)
                    else:
                        # Directory
                        if not return_root:
                            items += self.library_get_path_files_recursive(
                                data.path)
                        else:
                            items.append(data.path)
                else:
                    results, _playtime, _num_songs = \
                            self.library_return_search_items(
                                genre=data.genre, artist=data.artist, album=data.album,
                                year=data.year)
                    for item in results:
                        items.append(item.file)
        # Make sure we don't have any EXACT duplicates:
        items = misc.remove_list_duplicates(items, case=True)
        return items

    def library_get_path_files_recursive(self, path):
        results = []
        for item in self.mpd.lsinfo(path):
            if 'directory' in item:
                results = results + self.library_get_path_files_recursive(
                    item['directory'])
            elif 'file' in item:
                results.append(item['file'])
        return results

    def on_library_search_combo_change(self, _combo=None):
        self.config.last_search_num = self.searchcombo.get_active()
        if not self.search_visible():
            return
        self.prevlibtodo = ""
        self.prevlibtodo_base = "__"
        self.libsearchfilter_feed_loop(self.searchtext)

    def on_search_end(self, _button, move_focus=True):
        if self.search_visible():
            self.libsearchfilter_toggle(move_focus)

    def search_visible(self):
        return self.searchbutton.get_property('visible')

    def libsearchfilter_toggle(self, move_focus):
        if not self.search_visible() and self.connected():
            self.library.set_property('has-tooltip', True)
            ui.show(self.searchbutton)
            self.prevlibtodo = 'foo'
            self.prevlibtodo_base = "__"
            self.prevlibtodo_base_results = []
            # extra thread for background search work,
            # synchronized with a condition and its internal mutex
            self.libfilterbox_cond = threading.Condition()
            self.libfilterbox_cmd_buf = self.searchtext.get_text()
            qsearch_thread = threading.Thread(target=self.libsearchfilter_loop)
            qsearch_thread.name = "LibraryFilter"
            qsearch_thread.daemon = True
            qsearch_thread.start()
        elif self.search_visible():
            ui.hide(self.searchbutton)
            self.searchtext.handler_block(self.libfilter_changed_handler)
            self.searchtext.set_text("")
            self.searchtext.handler_unblock(self.libfilter_changed_handler)
            self.libsearchfilter_stop_loop()
            # call library_browse from the main thread to avoid corruption
            # of treeview, fixes #1959
            GLib.idle_add(self.library_browse, None, self.config.wd)
            if move_focus:
                self.library.grab_focus()

    def libsearchfilter_feed_loop(self, editable):
        if not self.search_visible():
            self.libsearchfilter_toggle(None)
        # Lets only trigger the searchfilter_loop if 200ms pass
        # without a change in Gtk.Entry
        try:
            GLib.source_remove(self.libfilterbox_source)
        except:
            pass
        self.libfilterbox_source = GLib.timeout_add(
            300, self.libsearchfilter_start_loop, editable)

    def libsearchfilter_start_loop(self, editable):
        self.libfilterbox_cond.acquire()
        self.libfilterbox_cmd_buf = editable.get_text()
        self.libfilterbox_cond.notifyAll()
        self.libfilterbox_cond.release()

    def libsearchfilter_stop_loop(self):
        self.libfilterbox_cond.acquire()
        self.libfilterbox_cmd_buf = '$$$QUIT###'
        self.libfilterbox_cond.notifyAll()
        self.libfilterbox_cond.release()

    def libsearchfilter_loop(self):
        while True:
            # copy the last command or pattern safely
            self.libfilterbox_cond.acquire()
            try:
                while(self.libfilterbox_cmd_buf == '$$$DONE###'):
                    self.libfilterbox_cond.wait()
                todo = self.libfilterbox_cmd_buf
                self.libfilterbox_cond.release()
            except:
                todo = self.libfilterbox_cmd_buf
            searchby = self.search_terms_mpd[self.config.last_search_num]
            if self.prevlibtodo != todo:
                if todo == '$$$QUIT###':
                    GLib.idle_add(ui.reset_entry_marking, self.searchtext)
                    return
                elif len(todo) > 1:
                    GLib.idle_add(self.libsearchfilter_do_search, searchby,
                                  todo)
                elif len(todo) == 0:
                    GLib.idle_add(ui.reset_entry_marking, self.searchtext)
                    self.libsearchfilter_toggle(False)
                else:
                    GLib.idle_add(ui.reset_entry_marking, self.searchtext)
            self.libfilterbox_cond.acquire()
            self.libfilterbox_cmd_buf = '$$$DONE###'
            try:
                self.libfilterbox_cond.release()
            except Exception as e:
                # XXX add logger here in the future!
                raise e
            self.prevlibtodo = todo

    def libsearchfilter_do_search(self, searchby, todo):
        if not self.prevlibtodo_base in todo:
            # Do library search based on first two letters:
            self.prevlibtodo_base = todo[:2]
            self.prevlibtodo_base_results = self.mpd.search(searchby,
                                                             self.prevlibtodo_base)
            subsearch = False
        else:
            subsearch = True

        # Now, use filtering similar to playlist filtering:
        # this make take some seconds... and we'll escape the search text
        # because we'll be searching for a match in items that are also escaped
        #
        # Note that the searching is not order specific. That is, "foo bar"
        # will match on "fools bar" and "barstool foo".

        todos = todo.split(" ")
        regexps = []
        for i in range(len(todos)):
            todos[i] = misc.escape_html(todos[i])
            todos[i] = re.escape(todos[i])
            todos[i] = '.*' + todos[i].lower()
            regexps.append(re.compile(todos[i]))
        matches = []
        if searchby != 'any':
            for row in self.prevlibtodo_base_results:
                is_match = True
                for regexp in regexps:
                    if not regexp.match(row.get(searchby, '').lower()):
                        is_match = False
                        break
                if is_match:
                    matches.append(row)
        else:
            for row in self.prevlibtodo_base_results:
                allstr = " ".join(row.values())
                is_match = True
                for regexp in regexps:
                    if not regexp.match(str(allstr).lower()):
                        is_match = False
                        break
                if is_match:
                    matches.append(row)
        if subsearch and len(matches) == len(self.librarydata):
            # nothing changed..
            return
        self.library.freeze_child_notify()
        currlen = len(self.librarydata)
        bd = [(self.sonatapb,
               SongRecord(path=item['file']),
               formatting.parse(self.config.libraryformat, item, True))
              for item in matches if 'file' in item]
        bd.sort(key=lambda key: locale.strxfrm(key[2]))
        for i, item in enumerate(bd):
            if i < currlen:
                j = self.librarydata.get_iter((i, ))
                for index in range(len(item)):
                    if item[index] != self.librarydata.get_value(j, index):
                        self.librarydata.set_value(j, index, item[index])
            else:
                self.librarydata.append(item)
        # Remove excess items...
        newlen = len(bd)
        if newlen == 0:
            self.librarydata.clear()
        else:
            for i in range(currlen - newlen):
                j = self.librarydata.get_iter((currlen - 1 - i,))
                self.librarydata.remove(j)
        self.library.thaw_child_notify()
        if len(matches) == 0:
            GLib.idle_add(ui.set_entry_invalid, self.searchtext)
        else:
            GLib.idle_add(self.library.set_cursor, Gtk.TreePath.new_first(),
                          None, False)
            GLib.idle_add(ui.reset_entry_marking, self.searchtext)

    def libsearchfilter_key_pressed(self, widget, event):
        self.filter_key_pressed(widget, event, self.library)

    def libsearchfilter_on_enter(self, _entry):
        self.on_library_row_activated(None, None)

    def libsearchfilter_set_focus(self):
        GLib.idle_add(self.searchtext.grab_focus)

    def libsearchfilter_get_style(self):
        return self.searchtext.get_style()



# Copyright 2006-2009 Scott Horowitz <stonecrest@gmail.com>
# Copyright 2009-2014 Jonathan Ballet <jon@multani.info>
#
# This file is part of Sonata.
#
# Sonata is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Sonata is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Sonata.  If not, see <http://www.gnu.org/licenses/>.

"""
This module implements a user interface for bookmarking remote music
streams.

Example usage:
import streams
self.streams = streams.Streams(self.config, self.window, self.on_streams_button_press, self.on_add_item, self.settings_save, self.TAB_STREAMS)
streamswindow, streamsevbox = self.streams.get_widgets()
...
self.streams.populate()
...
"""

import logging
import os

from gi.repository import Gtk, Gdk, Pango

from sonata import misc, ui


logger = logging.getLogger(__name__)


def parse_stream(stream_uri, fp):
    default = [stream_uri]

    # We only download the first 4k bytes, which should be sufficient to get an
    # idea if the stream is a M3U/PLS/etc. playlist or a binary file.
    content = fp.read(4000)
    try:
        content = content.decode()
        # If we can decode the first bytes, then read up to 64k bytes to parse
        # this as a playlist.
        content = content + fp.read(60000).decode()
    except UnicodeDecodeError: # Looks a binary content...
        logger.info("Adding binary stream from %s", stream_uri)
        return default

    if "[playlist]" in content:
        # pls:
        return parse_pls(content)
    elif "#EXTM3U" in content:
        # extended m3u:
        return parse_m3u(content)
    elif "http://" in content:
        # m3u or generic list:
        return parse_m3u(content)
    else:
        return default


def parse_pls(f):
    lines = f.split("\n")
    for line in lines:
        line = line.replace('\r', '')
        delim = line.find("=") + 1
        if delim > 0:
            line = line[delim:]
            if len(line) > 7 and line[0:7] == 'http://':
                yield line
            elif len(line) > 6 and line[0:6] == 'ftp://':
                yield line


def parse_m3u(f):
    lines = f.split("\n")
    for line in lines:
        line = line.replace('\r', '')
        if len(line) > 7 and line[0:7] == 'http://':
            yield line
        elif len(line) > 6 and line[0:6] == 'ftp://':
            yield line


class Streams:
    def __init__(self, config, window, on_streams_button_press, on_add_item, settings_save, TAB_STREAMS, add_tab):
        self.config = config
        self.window = window
        self.on_streams_button_press = on_streams_button_press
        self.on_add_item = on_add_item
        self.settings_save = settings_save

        self.stream_edit_dialog = None

        self.builder = ui.builder('streams')

        # Streams tab
        self.streams = self.builder.get_object('streams_page_treeview')
        self.streams_selection = self.streams.get_selection()
        self.streamswindow = self.builder.get_object('streams_page_scrolledwindow')

        self.tab_widget = self.builder.get_object('streams_tab_eventbox')
        self.tab_label = self.builder.get_object('streams_tab_label')
        self.tab_label.set_text(TAB_STREAMS)

        self.tab = add_tab(self.streamswindow, self.tab_widget, TAB_STREAMS,
                           self.streams)

        self.streams.connect('button_press_event', self.on_streams_button_press)
        self.streams.connect('row_activated', self.on_streams_activated)
        self.streams.connect('key-press-event', self.on_streams_key_press)

        # Initialize streams data and widget
        self.streamsdata = self.builder.get_object('streams_liststore')
        self.streams.set_search_column(1)

    def get_model(self):
        return self.streamsdata

    def get_widgets(self):
        return self.streamswindow

    def get_treeview(self):
        return self.streams

    def get_selection(self):
        return self.streams_selection

    def populate(self):
        self.streamsdata.clear()
        streamsinfo = [{'name' : misc.escape_html(name),
                'uri' : misc.escape_html(uri)}
                for name, uri in zip(self.config.stream_names,
                             self.config.stream_uris)]
        streamsinfo.sort(key=lambda x: x["name"].lower()) # Remove case sensitivity
        for item in streamsinfo:
            self.streamsdata.append([Gtk.STOCK_NETWORK, item["name"], item["uri"]])

    def on_streams_key_press(self, widget, event):
        if event.keyval == Gdk.keyval_from_name('Return'):
            self.on_streams_activated(widget, widget.get_cursor()[0])
            return True

    def on_streams_activated(self, _treeview, _path, _column=0):
        self.on_add_item(None)

    def on_streams_edit(self, action):
        model, selected = self.streams_selection.get_selected_rows()
        try:
            streamname = misc.unescape_html(model.get_value(model.get_iter(selected[0]), 1))
            for i, name in enumerate(self.config.stream_names):
                if name == streamname:
                    self.on_streams_new(action, i)
                    return
        except:
            pass

    def on_streams_new(self, _action, stream_num=-1):
        if stream_num > -1:
            edit_mode = True
        else:
            edit_mode = False
        if not self.stream_edit_dialog:
            self.stream_edit_dialog = self.builder.get_object(
                'stream_edit_dialog')
        self.stream_edit_dialog.set_transient_for(self.window)
        if edit_mode:
            self.stream_edit_dialog.set_title(_("Edit Stream"))
        else:
            self.stream_edit_dialog.set_title(_("New Stream"))
        nameentry = self.builder.get_object('stream_edit_name_entry')
        if edit_mode:
            nameentry.set_text(self.config.stream_names[stream_num])
        else:
            nameentry.set_text("")
        urlentry = self.builder.get_object('stream_edit_url_entry')
        if edit_mode:
            urlentry.set_text(self.config.stream_uris[stream_num])
        else:
            urlentry.set_text("")

        self.stream_edit_dialog.show_all()
        response = self.stream_edit_dialog.run()
        if response == Gtk.ResponseType.ACCEPT:
            name = nameentry.get_text()
            uri = urlentry.get_text()
            if len(name) > 0 and len(uri) > 0:
                # Make sure this stream name doesn't already exit:
                i = 0
                for item in self.config.stream_names:
                    # Prevent a name collision in edit_mode..
                    if not edit_mode or (edit_mode and i != stream_num):
                        if item == name:
                            self.stream_edit_dialog.hide()
                            if self._prompt_replace():
                                # Pop existing stream:
                                self.config.stream_names.pop(i)
                                self.config.stream_uris.pop(i)
                            else:
                                return
                    i = i + 1
                if edit_mode:
                    self.config.stream_names.pop(stream_num)
                    self.config.stream_uris.pop(stream_num)
                self.config.stream_names.append(name)
                self.config.stream_uris.append(uri)
                self.populate()
                self.settings_save()
        self.stream_edit_dialog.hide()

    def _prompt_replace(self):
        prompt = _("A stream with this name already exists. Would you like to replace it?")
        return ui.show_msg(self.window, prompt, _("New Stream"),
                           'newStreamError',
                           Gtk.ButtonsType.YES_NO) == Gtk.ResponseType.YES




# Copyright 2006-2009 Scott Horowitz <stonecrest@gmail.com>
# Copyright 2009-2014 Jonathan Ballet <jon@multani.info>
#
# This file is part of Sonata.
#
# Sonata is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Sonata is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Sonata.  If not, see <http://www.gnu.org/licenses/>.

"""
This module implements a user interface for mpd playlists.

Example usage:
import playlists
self.playlists = playlists.Playlists(self.config, self.window,
self.client, lambda:self.UIManager, self.update_menu_visibility,
self.iterate_now, self.on_add_item, self.on_playlists_button_press,
self.connected, self.TAB_PLAYLISTS)
playlistswindow, playlistsevbox = self.playlists.get_widgets()
...
self.playlists.populate()
...
"""

import os

from gi.repository import Gtk, Pango, Gdk

from sonata import ui, misc, mpdhelper as mpdh

from sonata.pluginsystem import pluginsystem, BuiltinPlugin


class Playlists:

    def __init__(self, config, window, mpd, UIManager,
                 update_menu_visibility, iterate_now, on_add_item,
                 on_playlists_button_press, connected,
                 add_selected_to_playlist, TAB_PLAYLISTS, add_tab):
        self.config = config
        self.window = window
        self.mpd = mpd
        self.UIManager = UIManager
        self.update_menu_visibility = update_menu_visibility
        self.iterate_now = iterate_now # XXX Do we really need this?
        self.on_add_item = on_add_item
        self.on_playlists_button_press = on_playlists_button_press
        self.add_selected_to_playlist = add_selected_to_playlist
        self.connected = connected

        self.mergepl_id = None
        self.actionGroupPlaylists = None
        self.playlist_name_dialog = None

        self.builder = ui.builder('playlists')

        # Playlists tab
        self.playlists = self.builder.get_object('playlists_page_treeview')
        self.playlists_selection = self.playlists.get_selection()
        self.playlistswindow = self.builder.get_object('playlists_page_scrolledwindow')

        self.tab_label = self.builder.get_object('playlists_tab_label')
        self.tab_label.set_text(TAB_PLAYLISTS)

        self.tab_widget = self.builder.get_object('playlists_tab_eventbox')
        self.tab = add_tab(self.playlistswindow, self.tab_widget, TAB_PLAYLISTS,
                           self.playlists)

        self.playlists.connect('button_press_event',
                               self.on_playlists_button_press)
        self.playlists.connect('row_activated', self.playlists_activated)
        self.playlists.connect('key-press-event', self.playlists_key_press)

        # Initialize playlist data and widget
        self.playlistsdata = self.builder.get_object('playlists_liststore')
        self.playlists.set_search_column(1)

    def get_model(self):
        return self.playlistsdata

    def get_widgets(self):
        return self.playlistswindow

    def get_treeview(self):
        return self.playlists

    def get_selection(self):
        return self.playlists_selection

    def populate_playlists_for_menu(self, playlistinfo):
        if self.mergepl_id:
            self.UIManager().remove_ui(self.mergepl_id)
        if self.actionGroupPlaylists:
            self.UIManager().remove_action_group(self.actionGroupPlaylists)
            self.actionGroupPlaylists = None
        self.actionGroupPlaylists = Gtk.ActionGroup('MPDPlaylists')
        self.UIManager().ensure_update()
        actions = [
            ("Playlist: %s" % playlist.replace("&", ""),
             Gtk.STOCK_JUSTIFY_CENTER,
             ui.quote_label(misc.unescape_html(playlist)),
             None, None,
             self.on_playlist_menu_click)
            for playlist in playlistinfo]
        self.actionGroupPlaylists.add_actions(actions)
        uiDescription = """
            <ui>
              <popup name="mainmenu">
                  <menu action="plmenu">
            """
        uiDescription += "".join('<menuitem action=\"%s\"/>' % action[0]
                        for action in actions)
        uiDescription += '</menu></popup></ui>'
        self.mergepl_id = self.UIManager().add_ui_from_string(uiDescription)
        self.UIManager().insert_action_group(self.actionGroupPlaylists, 0)
        self.UIManager().get_widget('/hidden').set_property('visible', False)
        # If we're not on the Current tab, prevent additional menu items
        # from displaying:
        self.update_menu_visibility()

    def on_playlist_save(self, _action):
        plname = self.prompt_for_playlist_name(_("Save Playlist"),
                                               'savePlaylist')
        if plname:
            if self.playlist_name_exists(_("Save Playlist"),
                                         'savePlaylistError', plname):
                return
            self.playlist_create(plname)
            self.mpd.playlistclear(plname)
            self.add_selected_to_playlist(plname)

    def playlist_create(self, playlistname, oldname=None):
        self.mpd.rm(playlistname)
        if oldname is not None:
            self.mpd.rename(oldname, playlistname)
        else:
            self.mpd.save(playlistname)
        self.populate()
        self.iterate_now()

    def on_playlist_menu_click(self, action):
        plname = misc.unescape_html(action.get_name().replace("Playlist: ",
                                                              ""))
        text = ('Would you like to replace the existing playlist or append'
                'these songs?')
        response = ui.show_msg(self.window,
                               _(text), _("Existing Playlist"),
                               "existingPlaylist", (_("Replace playlist"),
                                                    1, _("Append songs"), 2),
                               default=self.config.existing_playlist_option)
        if response == 1: # Overwrite
            self.config.existing_playlist_option = response
            self.mpd.playlistclear(plname)
            self.add_selected_to_playlist(plname)
        elif response == 2: # Append songs:
            self.config.existing_playlist_option = response
            self.add_selected_to_playlist(plname)

    def playlist_name_exists(self, title, role, plname, skip_plname=""):
        # If the playlist already exists, and the user does not want to
        # replace it, return True; In all other cases, return False
        playlists = self.mpd.listplaylists()
        if playlists is None:
            playlists = self.mpd.lsinfo()
        for item in playlists:
            if 'playlist' in item:
                if item['playlist'] == plname and \
                   plname != skip_plname:
                    if ui.show_msg(self.window,
                                   _(('A playlist with this name already '
                                     'exists. Would you like to replace it?')),
                                   title, role, Gtk.ButtonsType.YES_NO) == \
                       Gtk.ResponseType.YES:
                        return False
                    else:
                        return True
        return False

    def prompt_for_playlist_name(self, title, role, oldname=None):
        """Prompt user for playlist name"""
        plname = None
        if self.connected():
            if not self.playlist_name_dialog:
                self.playlist_name_dialog = self.builder.get_object(
                    'playlist_name_dialog')
            self.playlist_name_dialog.set_transient_for(self.window)
            self.playlist_name_dialog.set_title(title)
            self.playlist_name_dialog.set_role(role)
            entry = self.builder.get_object('playlist_name_entry')
            if oldname:
                entry.set_text(oldname)
            else:
                entry.set_text("")
            self.playlist_name_dialog.show_all()
            response = self.playlist_name_dialog.run()
            if response == Gtk.ResponseType.ACCEPT:
                plname = misc.strip_all_slashes(entry.get_text())
            self.playlist_name_dialog.hide()
        return plname

    def populate(self):
        if self.connected():
            self.playlistsdata.clear()
            playlistinfo = []
            playlists = self.mpd.listplaylists()
            if playlists is None:
                playlists = self.mpd.lsinfo()
            for item in playlists:
                if 'playlist' in item:
                    playlistinfo.append(misc.escape_html(item['playlist']))

            # Remove case sensitivity
            playlistinfo.sort(key=lambda x: x.lower())
            for item in playlistinfo:
                self.playlistsdata.append([Gtk.STOCK_DIRECTORY, item])

            self.populate_playlists_for_menu(playlistinfo)

    def on_playlist_rename(self, _action):
        model, selected = self.playlists_selection.get_selected_rows()
        oldname = misc.unescape_html(
            model.get_value(model.get_iter(selected[0]), 1))
        plname = self.prompt_for_playlist_name(_("Rename Playlist"),
                                               'renamePlaylist', oldname)
        if plname:
            if self.playlist_name_exists(_("Rename Playlist"),
                                         'renamePlaylistError',
                                         plname, oldname):
                return
            self.playlist_create(plname, oldname)
            # Re-select item:
            row = 0
            for pl in self.playlistsdata:
                if pl[1] == plname:
                    self.playlists_selection.select_path((row,))
                    return
                row = row + 1

    def playlists_key_press(self, widget, event):
        if event.keyval == Gdk.keyval_from_name('Return'):
            self.playlists_activated(widget, widget.get_cursor()[0])
            return True

    def playlists_activated(self, _treeview, _path, _column=0):
        self.on_add_item(None)
